from starlette.requests import Request
from starlette.responses import Response

from spinta import commands
from spinta.accesslog import AccessLog
from spinta.renderer import render
from spinta.components import Context, Property, Action, UrlParams
from spinta.backends.components import BackendFeatures
from spinta.backends.postgresql.files import DatabaseFile
from spinta.utils.data import take
from spinta.types.datatype import File
from spinta.exceptions import NotFoundError, ItemDoesNotExist
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.constants import TableType
from spinta.utils.nestedstruct import flat_dicts_to_nested


@commands.getone.register(Context, Request, Property, File, PostgreSQL)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: File,
    backend: PostgreSQL,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, prop)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.log(
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
    )
    data = getone(context, prop, dtype, backend, id_=params.pk)

    # Return file metadata
    if params.propref:
        data = commands.prepare_data_for_response(
            context,
            prop.dtype,
            params.fmt,
            data,
            action=Action.GETONE,
        )
        return render(context, request, prop, params, data, action=action)

    # Return file content
    else:
        value = take(prop.place, data)

        if not take('_blocks', value):
            raise ItemDoesNotExist(dtype, id=params.pk)

        filename = value['_id']

        connection = context.get('transaction').connection
        table = backend.get_table(prop, TableType.FILE)
        with DatabaseFile(
            connection,
            table,
            value['_size'],
            value['_blocks'],
            value['_bsize'],
            mode='r',
        ) as f:
            content = f.read()

        return Response(
            content,
            media_type=value['_content_type'],
            headers={
                'Revision': data['_revision'],
                'Content-Disposition': (
                    f'attachment; filename="{filename}"'
                    if filename else
                    'attachment'
                )
            },
        )


@commands.getone.register(Context, Property, File, PostgreSQL)
def getone(
    context: Context,
    prop: Property,
    dtype: File,
    backend: PostgreSQL,
    *,
    id_: str,
):
    table = backend.get_table(prop.model)
    connection = context.get('transaction').connection
    selectlist = [
        table.c._id,
        table.c._revision,
        table.c[prop.place + '._id'],
        table.c[prop.place + '._content_type'],
        table.c[prop.place + '._size'],
    ]

    if BackendFeatures.FILE_BLOCKS in prop.dtype.backend.features:
        selectlist += [
            table.c[prop.place + '._bsize'],
            table.c[prop.place + '._blocks'],
        ]

    try:
        data = backend.get(connection, selectlist, table.c._id == id_)
    except NotFoundError:
        raise ItemDoesNotExist(dtype, id=id_)

    result = {
        '_type': prop.model_type(),
        '_id': data[table.c._id],
        '_revision': data[table.c._revision],
    }

    data = flat_dicts_to_nested(data)
    result[prop.name] = data[prop.name]
    return commands.cast_backend_to_python(context, prop, backend, result)


@commands.getfile.register()
def getfile(
    context: Context,
    prop: Property,
    dtype: File,
    backend: PostgreSQL,
    *,
    data: dict,
):
    if not data['_blocks']:
        return None

    if len(data['_blocks']) > 1:
        # TODO: Use propper UserError exception.
        raise Exception(
            "File content is to large to retrun it inline. Try accessing "
            "this file directly using subresources API."
        )

    connection = context.get('transaction').connection
    table = backend.get_table(prop, TableType.FILE)
    with DatabaseFile(
        connection,
        table,
        data['_size'],
        data['_blocks'],
        data['_bsize'],
        mode='r',
    ) as f:
        return f.read()



import dataclasses
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from spinta.formats.components import Format


class Html(Format):
    content_type = 'text/html'
    accept_types = {
        'text/html',
    }
    params = {}
    streamable = False


class Color(str, Enum):
    change = '#B2E2AD'
    null = '#f5f5f5'


@dataclasses.dataclass
class Cell:
    value: str
    link: Optional[str] = None
    color: Optional[Color] = None

    def as_dict(self) -> Dict[str, Any]:
        data = {
            'value': self.value,
            'link': self.link,
            'color': str(self.color.value) if self.color else None,
        }
        return {
            k: v
            for k, v in data.items()
            if k == 'value' or v is not None
        }


ComplexCell = Union[
    Cell,
    Dict[str, 'ComplexCell'],
    List['ComplexCell'],
]





from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.backends.memory.components import Memory


@commands.prepare.register(Context, Memory, Manifest)
def prepare(context: Context, backend: Memory, manifest: Manifest):
    pass



from typing import Any

from spinta.core.ufuncs import Env
from spinta.core.ufuncs import ufunc


@ufunc.resolver(Env, object, object)
def swap(env: Env, old: Any, new: Any) -> Any:
    return env.call('swap', env.this, old, new)


@ufunc.resolver(Env, object, object, object)
def swap(env: Env, this: Any, old: Any, new: Any) -> Any:
    if this == old:
        return new
    else:
        return this



from spinta import commands
from spinta.components import Context
from spinta.backends.mongo.components import Mongo


@commands.wait.register(Context, Mongo)
def wait(context: Context, backend: Mongo, *, fail: bool = False) -> bool:
    # TODO: Implement real check.
    return True



import uuid
import csv
import urllib.parse
from typing import Dict, Any

from spinta import commands
from spinta.fetcher import fetch
from spinta.core.ufuncs import Expr
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.datasets.components import Entity
from spinta.datasets.utils import iterparams
from spinta.datasets.backends.csv.components import Csv


@commands.load.register(Context, Csv, dict)
def load(context: Context, backend: Csv, config: Dict[str, Any]):
    pass


@commands.prepare.register(Context, Csv, Manifest)
def prepare(context: Context, backend: Csv, manifest: Manifest):
    pass


@commands.bootstrap.register(Context, Csv)
def bootstrap(context: Context, backend: Csv):
    pass


@commands.getall.register(Context, Entity, Csv)
def getall(
    context: Context,
    external: Entity,
    backend: Csv,
    *,
    query: Expr = None,
):
    model = external.model
    base = external.resource.external
    props = {
        p.external.name: p.name
        for p in model.properties.values()
        if p.external
    }
    for params in iterparams(context, model):
        url = urllib.parse.urljoin(base, external.name) if base else external.name
        url = url.format(**params)
        with fetch(context, url, text=True).open() as f:
            for row in csv.DictReader(f):
                row = {
                    params[k]: v
                    for k, v in row.items()
                    if k in props
                }
                row['_type'] = model.model_type()
                row['_id'] = str(uuid.uuid4())
                yield commands.cast_backend_to_python(context, model, backend, row)



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

from spinta import exceptions
from spinta.exceptions import BackendNotFound

if TYPE_CHECKING:
    from spinta.types.datatype import DataType


def check_no_extra_keys(dtype: DataType, schema: Iterable, data: Iterable):
    unknown = set(data) - set(schema)
    if unknown:
        raise exceptions.MultipleErrors(
            exceptions.FieldNotInResource(
                dtype.prop,
                property=f'{dtype.prop.place}.{prop}',
            )
            for prop in sorted(unknown)
        )


def set_dtype_backend(dtype: DataType):
    if dtype.backend:
        backends = dtype.prop.model.manifest.store.backends
        if dtype.backend not in backends:
            raise BackendNotFound(dtype, name=dtype.backend)
        dtype.backend = backends[dtype.backend]
    else:
        dtype.backend = dtype.prop.model.backend



import pathlib
from typing import Optional

from spinta import commands
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.exceptions import ManifestFileDoesNotExist
from spinta.manifests.tabular.components import AsciiManifest
from spinta.manifests.tabular.components import CsvManifest
from spinta.manifests.tabular.components import TabularManifest


@commands.configure.register(Context, TabularManifest)
def configure(context: Context, manifest: TabularManifest):
    rc: RawConfig = context.get('rc')
    path: Optional[str] = rc.get('manifests', manifest.name, 'path')
    if path:
        if (
            not path.startswith(('http://', 'https://')) and
            not pathlib.Path(path).exists()
        ):
            raise ManifestFileDoesNotExist(manifest, path=path)
    else:
        path = None
    manifest.path = path

    if isinstance(manifest, (CsvManifest, AsciiManifest)):
        manifest.file = rc.get('manifests', manifest.name, 'file')



from typing import List
from typing import Optional
from typing import Tuple

from spinta import exceptions
from spinta.auth import authorized
from spinta.backends import Backend
from spinta.backends.components import BackendOrigin
from spinta.backends.helpers import load_backend
from spinta.components import Action
from spinta.components import Context
from spinta.components import Model
from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import ShortExpr
from spinta.datasets.components import Resource
from spinta.dimensions.enum.helpers import get_prop_enum
from spinta.types.datatype import Ref
from spinta.ufuncs.changebase.helpers import change_base_model
from spinta.ufuncs.components import ForeignProperty
from spinta.ufuncs.helpers import merge_formulas
from spinta.utils.data import take
from spinta.utils.naming import to_code_name


def load_resource_backend(
    context: Context,
    resource: Resource,
    name: Optional[str],  # Backend name given in `ref` column.
) -> Backend:
    if name:
        # If backend name is given, just use an exiting backend.
        store = resource.dataset.manifest.store
        possible_backends = [
            # Backend defined in manifest.
            resource.dataset.manifest.backends,
            # Backend defined via configuration.
            store.backends,
        ]
        for backends in possible_backends:
            if name in backends:
                return backends[name]
        raise exceptions.BackendNotFound(resource, name=name)

    elif resource.type or resource.external:
        # If backend is defined inline on resource itself, try to load it.
        name = f'{resource.dataset.name}/{resource.name}'
        return load_backend(context, resource, name, BackendOrigin.resource, {
            'type': resource.type,
            'name': name,
            'dsn': resource.external,
        })

    else:
        # If backend was not given, try to check configuration with resource and
        # dataset names as backend names or fallback to manifest backend.
        store = resource.dataset.manifest.store
        possible_backends = [
            # Dataset and resource name combination.
            f'{resource.dataset.name}/{resource.name}',
            # Just dataset name.
            resource.dataset.name,
            # Just resource type.
            # XXX: I don't think that using resource type as a backend name is a
            #      good idea. Probably should be removed.
            resource.type,
        ]
        for name in filter(None, possible_backends):
            name = to_code_name(name)
            if name in store.backends:
                return store.backends[name]
        return resource.dataset.manifest.backend


def get_enum_filters(
    context: Context,
    model: Model,
    fpr: ForeignProperty = None,
) -> Optional[Expr]:
    args: List[Expr] = []
    for prop in take(['_id', all], model.properties).values():
        if (
            (enum := get_prop_enum(prop)) and
            authorized(context, prop, Action.GETALL)
        ):
            if not all(item.access >= prop.access for item in enum.values()):
                values = []
                for item in enum.values():
                    if item.access >= prop.access:
                        values.append(item.prepare)

                if fpr:
                    dtype = fpr.push(prop)
                else:
                    dtype = prop.dtype

                if len(values) == 1:
                    arg = values[0]
                elif len(values) > 1:
                    arg = values
                else:
                    arg = None

                if arg is not None:
                    expr = ShortExpr('eq', dtype.get_bind_expr(), arg)
                    args.append(expr)
    if len(args) == 1:
        return args[0]
    elif len(args) > 1:
        return ShortExpr('and', *args)
    else:
        return None


def get_ref_filters(
    context: Context,
    model: Model,
    fpr: ForeignProperty = None,
    *,
    seen: List[Tuple[
        str,  # absolute model name
        str,  # model property
    ]] = None,
) -> Optional[Expr]:
    query: Optional[Expr] = None
    if fpr:
        query = merge_formulas(
            query,
            change_base_model(context, model, fpr),
        )
        query = merge_formulas(
            query,
            get_enum_filters(context, model, fpr),
        )

    seen = seen or []

    for prop in take(['_id', all], model.properties).values():
        ref_key = (model.name, prop.name)
        if (
            isinstance(prop.dtype, Ref) and
            ref_key not in seen and
            authorized(context, prop, Action.GETALL)
        ):
            if fpr:
                _fpr = fpr.push(prop)
            else:
                _fpr = ForeignProperty(fpr, prop.dtype)

            query = merge_formulas(query, get_ref_filters(
                context,
                prop.dtype.model,
                _fpr,
                seen=seen + [ref_key],
            ))

    return query



import datetime


from spinta import commands
from spinta.utils.data import take
from spinta.components import Context, Model, DataStream, DataItem, DataSubItem
from spinta.exceptions import ItemDoesNotExist
from spinta.types.datatype import DataType, Ref
from spinta.backends.mongo.components import Mongo
from spinta.backends.mongo.helpers import inserting


@commands.insert.register()
async def insert(
    context: Context,
    model: Model,
    backend: Mongo,
    *,
    dstream: DataStream,
    stop_on_error: bool = True,
):
    table = backend.db[model.model_type()]
    async for data in dstream:
        patch = commands.before_write(context, model, backend, data=data)
        # TODO: Insert batches in a single query, using `insert_many`.
        table.insert_one(patch)
        commands.after_write(context, model, backend, data=data)
        yield data


@commands.update.register()
async def update(
    context: Context,
    model: Model,
    backend: Mongo,
    *,
    dstream: dict,
    stop_on_error: bool = True,
):
    table = backend.db[model.model_type()]
    async for data in dstream:
        patch = commands.before_write(context, model, backend, data=data)
        result = table.update_one(
            {
                '__id': data.saved['_id'],
                '_revision': data.saved['_revision'],
            },
            {'$set': patch},
        )
        if result.matched_count == 0:
            raise ItemDoesNotExist(
                model,
                id=data.saved['_id'],
                revision=data.saved['_revision'],
            )
        assert result.matched_count == 1 and result.modified_count == 1, (
            f"matched: {result.matched_count}, modified: {result.modified_count}"
        )
        commands.after_write(context, model, backend, data=data)
        yield data


@commands.delete.register()
async def delete(
    context: Context,
    model: Model,
    backend: Mongo,
    *,
    dstream: dict,
    stop_on_error: bool = True,
):
    table = backend.db[model.model_type()]
    async for data in dstream:
        commands.before_write(context, model, backend, data=data)
        result = table.delete_one({
            '__id': data.saved['_id'],
            '_revision': data.saved['_revision'],
        })
        if result.deleted_count == 0:
            # FIXME: Respect stop_on_error flag.
            raise ItemDoesNotExist(
                model,
                id=data.saved['_id'],
                revision=data.saved['_revision'],
            )
        commands.after_write(context, model, backend, data=data)
        yield data


@commands.before_write.register(Context, Model, Mongo)
def before_write(
    context: Context,
    model: Model,
    backend: Mongo,
    *,
    data: DataItem,
) -> dict:
    patch = {}
    patch['__id'] = take('_id', data.patch)
    patch['_revision'] = take('_revision', data.patch, data.saved)
    patch['_txn'] = context.get('transaction').id
    patch['_created'] = datetime.datetime.now(datetime.timezone.utc)
    for prop in take(model.properties).values():
        value = commands.before_write(
            context,
            prop.dtype,
            backend,
            data=data[prop.name],
        )
        patch.update(value)
    return take(all, patch)


@commands.before_write.register(Context, DataType, Mongo)
def before_write(
    context: Context,
    dtype: DataType,
    backend: Mongo,
    *,
    data: DataSubItem,
) -> dict:
    if inserting(data):
        return {dtype.prop.name: data.patch}
    else:
        return {dtype.prop.place: data.patch}


@commands.before_write.register(Context, Ref, Mongo)
def before_write(
    context: Context,
    dtype: Ref,
    backend: Mongo,
    *,
    data: DataSubItem,
) -> dict:
    patch = take(['_id'], data.patch) or None
    if inserting(data):
        return {dtype.prop.name: patch}
    else:
        return {dtype.prop.place: patch}


@commands.after_write.register(Context, Model, Mongo)
def after_write(
    context: Context,
    model: Model,
    backend: Mongo,
    *,
    data: DataItem,
) -> dict:
    pass



from typing import AsyncIterator


from spinta import commands
from spinta.utils.data import take
from spinta.components import Context, Model, DataItem, DataSubItem
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.sqlalchemy import utcnow


@commands.insert.register(Context, Model, PostgreSQL)
async def insert(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    transaction = context.get('transaction')
    connection = transaction.connection
    table = backend.get_table(model)
    async for data in dstream:
        patch = commands.before_write(context, model, backend, data=data)

        # TODO: Refactor this to insert batches with single query.
        qry = table.insert().values(
            _id=patch['_id'],
            _revision=patch['_revision'],
            _txn=transaction.id,
            _created=utcnow(),
        )
        connection.execute(qry, patch)

        commands.after_write(context, model, backend, data=data)

        yield data


@commands.update.register(Context, Model, PostgreSQL)
async def update(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    dstream: dict,
    stop_on_error: bool = True,
):
    transaction = context.get('transaction')
    connection = transaction.connection
    table = backend.get_table(model)

    async for data in dstream:
        if not data.patch:
            yield data
            continue

        pk = data.saved['_id']
        patch = commands.before_write(context, model, backend, data=data)
        result = connection.execute(
            table.update().
            where(table.c._id == pk).
            where(table.c._revision == data.saved['_revision']).
            values(patch)
        )

        if result.rowcount == 0:
            raise Exception(f"Update failed, {model} with {pk} not found.")
        elif result.rowcount > 1:
            raise Exception(
                f"Update failed, {model} with {pk} has found and update "
                f"{result.rowcount} rows."
            )

        commands.after_write(context, model, backend, data=data)

        yield data


@commands.delete.register(Context, Model, PostgreSQL)
async def delete(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    transaction = context.get('transaction')
    connection = transaction.connection
    table = backend.get_table(model)
    async for data in dstream:
        commands.before_write(context, model, backend, data=data)
        connection.execute(
            table.delete().
            where(table.c._id == data.saved['_id'])
        )
        commands.after_write(context, model, backend, data=data)
        yield data


@commands.before_write.register(Context, Model, PostgreSQL)
def before_write(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    data: DataSubItem,
) -> dict:
    patch = take(['_id'], data.patch)
    patch['_revision'] = take('_revision', data.patch, data.saved)
    patch['_txn'] = context.get('transaction').id
    patch['_created'] = utcnow()
    for prop in take(model.properties).values():
        value = commands.before_write(
            context,
            prop.dtype,
            backend,
            data=data[prop.name],
        )
        patch.update(value)
    return patch


@commands.after_write.register(Context, Model, PostgreSQL)
def after_write(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    data: DataSubItem,
) -> dict:
    for key in take(data.patch or {}):
        prop = model.properties[key]
        commands.after_write(context, prop.dtype, backend, data=data[key])



from spinta import commands
from spinta.components import Context
from spinta.components import Model
from spinta.backends.memory.components import Memory


@commands.wipe.register(Context, Model, Memory)
def wipe(context: Context, model: Model, backend: Memory):
    if model.name in backend.data:
        backend.data[model.name] = {}



import sqlalchemy as sa

from sqlalchemy.dialects.postgresql import JSONB

from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Array
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.constants import TableType
from spinta.backends.postgresql.helpers import get_pg_name, get_table_name


@commands.prepare.register(Context, PostgreSQL, Array)
def prepare(context: Context, backend: PostgreSQL, dtype: Array):
    prop = dtype.prop

    columns = commands.prepare(context, backend, dtype.items)
    assert columns is not None
    if not isinstance(columns, list):
        columns = [columns]

    pkey_type = commands.get_primary_key_type(context, backend)

    # TODO: When all list items will have unique id, also add reference to
    #       parent list id.
    # if prop.list:
    #     parent_list_table_name = get_pg_name(
    #         get_table_name(prop.list, TableType.LIST)
    #     )
    #     columns += [
    #         # Parent list table id.
    #         sa.Column('_lid', pkey_type, sa.ForeignKey(
    #             f'{parent_list_table_name}._id', ondelete='CASCADE',
    #         )),
    #     ]

    name = get_pg_name(get_table_name(prop, TableType.LIST))
    main_table_name = get_pg_name(get_table_name(prop.model))
    table = sa.Table(
        name, backend.schema,
        # TODO: List tables eventually will have _id in order to uniquelly
        #       identify list item.
        # sa.Column('_id', pkey_type, primary_key=True),
        sa.Column('_txn', pkey_type, index=True),
        # Main table id (resource id).
        sa.Column('_rid', pkey_type, sa.ForeignKey(
            f'{main_table_name}._id', ondelete='CASCADE',
        )),
        *columns,
    )
    backend.add_table(table, prop, TableType.LIST)

    if prop.list is None:
        # For fast whole resource access we also store whole list in a JSONB.
        return sa.Column(prop.place, JSONB)



import sqlalchemy as sa

from sqlalchemy.dialects.postgresql import BIGINT, ARRAY

from spinta import commands
from spinta.components import Context
from spinta.types.datatype import File
from spinta.backends.postgresql.components import PostgreSQL, BackendFeatures
from spinta.backends.postgresql.constants import TableType
from spinta.backends.postgresql.helpers import get_pg_name, get_table_name
from spinta.backends.postgresql.helpers import get_column_name


@commands.prepare.register(Context, PostgreSQL, File)
def prepare(context: Context, backend: PostgreSQL, dtype: File):
    prop = dtype.prop
    model = prop.model

    # All properties receive model backend even if properties own backend
    # differs. Its properties responsibility to deal with foreign backends.
    assert model.backend.name == backend.name

    pkey_type = commands.get_primary_key_type(context, model.backend)

    same_backend = backend.name == dtype.backend.name

    if same_backend:
        # If dtype is on the same backend currently being prepared, then
        # create table for storing file blocks and also add file metadata
        # columns.
        table_name = get_pg_name(get_table_name(prop, TableType.FILE))
        table = sa.Table(
            table_name, model.backend.schema,
            sa.Column('_id', pkey_type, primary_key=True),
            sa.Column('_block', sa.LargeBinary),
        )
        model.backend.add_table(table, prop, TableType.FILE)

    name = get_column_name(prop)

    # Required file metadata on any backend.
    columns = [
        sa.Column(f'{name}._id', sa.String),
        sa.Column(f'{name}._content_type', sa.String),
        sa.Column(f'{name}._size', BIGINT),
    ]

    # Optional file metadata, depending on backend supported features.
    if same_backend or BackendFeatures.FILE_BLOCKS in dtype.backend.features:
        columns += [
            sa.Column(f'{name}._bsize', sa.Integer),
            sa.Column(f'{name}._blocks', ARRAY(pkey_type)),
        ]

    return columns



from spinta import commands
from spinta.components import Context
from spinta.manifests.yaml.components import YamlManifest


@commands.sync.register()
def sync(context: Context, manifest: YamlManifest):
    if manifest.sync:
        # TODO: sync YAML files from other manifests
        raise NotImplementedError



import enum


# Maximum length for PostgreSQL identifiers (e.g. table names, column names,
# function names).
# https://github.com/postgres/postgres/blob/master/src/include/pg_config_manual.h
NAMEDATALEN = 63

UNSUPPORTED_TYPES = [
    'backref',
    'generic',
    'rql',
]


class TableType(enum.Enum):
    MAIN = ''
    LIST = '/:list'
    CHANGELOG = '/:changelog'
    CACHE = '/:cache'
    FILE = '/:file'



import contextlib

import requests

from spinta.datasets.components import ExternalBackend


class Csv(ExternalBackend):

    @contextlib.contextmanager
    def begin(self):
        yield requests.Session()



import asyncio

from spinta import commands
from spinta.components import Context
from spinta.manifests.backend.components import BackendManifest
from spinta.manifests.backend.helpers import run_bootstrap


@commands.bootstrap.register(Context, BackendManifest)
def bootstrap(context: Context, manifest: BackendManifest):
    backend = manifest.backend
    if not backend.bootstrapped():
        with context:
            context.attach('transaction', manifest.backend.transaction, write=True)
            loop = asyncio.get_event_loop()
            loop.run_until_complete(run_bootstrap(context, manifest))



from starlette.requests import Request

from spinta import commands
from spinta.backends.helpers import get_select_prop_names
from spinta.backends.helpers import get_select_tree
from spinta.compat import urlparams_to_expr
from spinta.components import Context, Node, Action, UrlParams
from spinta.components import Model
from spinta.components import Property
from spinta.datasets.components import ExternalBackend
from spinta.renderer import render
from spinta.types.datatype import Integer


@commands.getall.register(Context, Model, Request)
async def getall(
    context: Context,
    model: Model,
    request: Request,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, model)

    backend = model.backend

    if isinstance(backend, ExternalBackend):
        # XXX: `add_count` is a hack, because, external backends do not
        #      support it yet.
        expr = urlparams_to_expr(params, add_count=False)
    else:
        expr = urlparams_to_expr(params)

    accesslog = context.get('accesslog')
    accesslog.log(
        model=model.model_type(),
        action=action.value,
    )

    rows = commands.getall(context, model, backend, query=expr)

    if params.count:
        # XXX: Quick and dirty hack. Functions should be handled properly.
        prop = Property()
        prop.name = 'count()'
        prop.place = 'count()'
        prop.title = ''
        prop.description = ''
        prop.model = model
        prop.dtype = Integer()
        prop.dtype.type = 'integer'
        prop.dtype.type_args = []
        prop.dtype.name = 'integer'
        prop.dtype.prop = prop
        props = {
            '_type': model.properties['_type'],
            'count()': prop,
        }
        rows = (
            {
                prop.name: commands.prepare_dtype_for_response(
                    context,
                    props[key].dtype,
                    params.fmt,
                    val,
                    data=row,
                    action=action,
                )
                for key, val in row.items()
            }
            for row in rows
        )
    else:
        select_tree = get_select_tree(context, action, params.select)
        prop_names = get_select_prop_names(
            context,
            model,
            action,
            select_tree,
            reserved=['_type', '_id', '_revision'],
        )
        rows = (
            commands.prepare_data_for_response(
                context,
                model,
                params.fmt,
                row,
                action=action,
                select=select_tree,
                prop_names=prop_names,
            )
            for row in rows
        )
    return render(context, request, model, params, rows, action=action)


@commands.getone.register(Context, Request, Node)
async def getone(
    context: Context,
    request: Request,
    node: Node,
    *,
    action: Action,
    params: UrlParams,
):
    if params.prop and params.propref:
        return await commands.getone(context, request, params.prop, params.model.backend, action=action, params=params)
    elif params.prop:
        return await commands.getone(context, request, params.prop, params.prop.dtype.backend or params.model.backend, action=action, params=params)
    else:
        return await commands.getone(context, request, params.model, params.model.backend, action=action, params=params)


@commands.changes.register(Context, Model, Request)
async def changes(
    context: Context,
    model: Model,
    request: Request,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, model)
    rows = commands.changes(
        context,
        model,
        model.backend,
        id_=params.pk,
        limit=params.limit,
        offset=params.offset,
    )
    select_tree = get_select_tree(context, action, params.select)
    prop_names = get_select_prop_names(
        context,
        model,
        action,
        select_tree,
        reserved=[
            '_cid',
            '_created',
            '_op',
            '_id',
            '_txn',
            '_revision',
        ],
    )
    rows = (
        commands.prepare_data_for_response(
            context,
            model,
            params.fmt,
            row,
            action=action,
            select=select_tree,
            prop_names=prop_names,
        )
        for row in rows
    )
    return render(context, request, model, params, rows, action=action)



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

import contextlib

from pytest import FixtureRequest

from spinta import commands
from spinta.cli.helpers.store import prepare_manifest
from spinta.components import Node
from spinta.components import Context
from spinta.auth import AdminToken
from spinta.utils.imports import importstr
from spinta.core.context import create_context
from spinta.core.config import RawConfig


def create_test_context(
    rc: RawConfig,
    request: FixtureRequest = None,
    *,
    name: str = 'pytest',
) -> TestContext:
    rc = rc.fork()
    Context_ = rc.get('components', 'core', 'context', cast=importstr)
    Context_ = type('ContextForTests', (ContextForTests, Context_), {})
    context = Context_(name)
    context = create_context(name, rc, context)
    if request:
        request.addfinalizer(context.wipe_all)
    return context


class ContextForTests:
    loaded: bool = False

    def __init__(
        self: TestContext,
        name: str,
        parent: TestContext = None,
    ):
        super().__init__(name, parent)
        self.loaded = parent.loaded if parent else False

    @contextlib.contextmanager
    def transaction(self: TestContext, *, write=False):
        if self.has('transaction'):
            yield self
        else:
            with self:
                store = self.get('store')
                self.set('auth.token', AdminToken())
                backend = store.manifest.backend
                self.attach('transaction', backend.transaction, write=write)
                yield self

    def wipe(self: TestContext, model: Union[str, Node]):
        if isinstance(model, str):
            store = self.get('store')
            model = store.manifest.models[model]
        with self.transaction() as context:
            commands.wipe(context, model, model.backend)

    def wipe_all(self: TestContext):
        store = self.get('store')
        self.wipe(store.manifest.objects['ns'][''])

    def load(
        self: TestContext,
        overrides: Optional[Dict[str, Any]] = None,
    ) -> TestContext:
        # We pass context to tests unloaded, by doing this, we give test
        # functions opportunity to call `context.load` manually and provide
        # `overrides` for config, this way each test can configure context in
        # anyway they want.
        #
        # If test function does not explicitly call `context.load`, then it will
        # be called implicitly on `app.request` and on some context methods,
        # that run database queries.
        if self.loaded:
            raise Exception("test context is already loaded")

        rc: RawConfig = self.get('rc')

        if overrides:
            rc.add('test', {
                'environments': {
                    'test': overrides,
                }
            })

        store = prepare_manifest(self)
        commands.bootstrap(self, store.manifest)

        self.loaded = True

        return self


TestContext = Union[Context, ContextForTests]



import contextlib

from spinta.backends.components import Backend
from spinta.backends.components import BackendFeatures


class Memory(Backend):
    data: dict = None

    features = {
        BackendFeatures.WRITE,
    }

    @contextlib.contextmanager
    def begin(self):
        yield self

    def bootstrapped(self):
        return True



import datetime
import types

from spinta import commands
from spinta.utils.json import fix_data_for_json
from spinta.components import Context, Model, Property
from spinta.backends.mongo.components import Mongo


@commands.create_changelog_entry.register(Context, (Model, Property), Mongo)
async def create_changelog_entry(
    context: Context,
    model: (Model, Property),
    backend: Mongo,
    *,
    dstream: types.AsyncGeneratorType,
) -> None:
    transaction = context.get('transaction')
    if isinstance(model, Model):
        table = backend.db[model.model_type() + '__changelog']
    else:
        table = backend.db[model.model.model_type() + '__changelog']
    async for data in dstream:
        table.insert_one({
            '__id': data.saved['_id'] if data.saved else data.patch['_id'],
            '_revision': data.patch['_revision'] if data.patch else data.saved['_revision'],
            '_op': data.action.value,
            '_transaction': transaction.id,
            '_created': datetime.datetime.now(datetime.timezone.utc),
            **fix_data_for_json({
                k: v for k, v in data.patch.items() if not k.startswith('_')
            }),
        })
        yield data



from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.manifests.backend.components import BackendManifest


@commands.load.register(Context, BackendManifest)
def load(
    context: Context,
    manifest: BackendManifest,
    *,
    into: Manifest = None,
    freezed: bool = True,
    rename_duplicates: bool = False,
    load_internal: bool = True,
):
    assert freezed, (
        "BackendManifest does not have unfreezed version of manifest."
    )
    commands.load(context, manifest, manifest.backend, into=into, freezed=freezed)



from __future__ import annotations

import logging
import pathlib
from typing import List
from typing import Optional

from typer import Context as TyperContext
from typer import Option
from typer import Typer
from typer import echo

import spinta
from spinta.cli import auth
from spinta.cli import config
from spinta.cli import data
from spinta.cli import inspect
from spinta.cli import manifest
from spinta.cli import migrate
from spinta.cli import pii
from spinta.cli import pull
from spinta.cli import push
from spinta.cli import server
from spinta.cli.init import init
from spinta.cli.show import show
from spinta.cli.helpers.typer import add
from spinta.core.context import create_context

log = logging.getLogger(__name__)

app = Typer()

add(app, 'config', config.config, short_help="Show current configuration values")
add(app, 'check', config.check, short_help="Check configuration and manifests")

add(app, 'genkeys', auth.genkeys, short_help="Generate client token validation keys")
add(app, 'token', auth.token, short_help="Tools for encoding/decode auth tokens")
add(app, 'client', auth.client, short_help="Manage auth clients")

add(app, 'init', init, short_help="Initialize a new manifest table")
add(app, 'inspect', inspect.inspect, short_help=(
    "Update manifest schema from an external data source"
))
add(app, 'pii', pii.app, short_help="Manage Person Identifying Information")

add(app, 'copy', manifest.copy, short_help="Copy only specified metadata from a manifest")
add(app, 'show', show, short_help="Show manifest as ascii table")

add(app, 'bootstrap', migrate.bootstrap, short_help="Initialize backends")
add(app, 'sync', migrate.sync, short_help="Sync source manifests into main manifest")
add(app, 'migrate', migrate.migrate, short_help="Migrate schema changes to backends")
add(app, 'freeze', migrate.freeze, short_help=(
    "Detect schema changes and create new schema version"
))

add(app, 'import', data.import_, short_help="Import data from a file")
add(app, 'pull', pull.pull, short_help="Pull data from an external data source")
add(app, 'push', push.push, short_help="Push data into an external data store")

add(app, 'run', server.run, short_help="Run development server")
add(app, 'wait', server.wait, short_help="Wait while all backends are up")


@app.callback(invoke_without_command=True)
def main(
    ctx: TyperContext,
    option: Optional[List[str]] = Option(None, '-o', '--option', help=(
        "Set configuration option, example: `-o option.name=value`."
    )),
    env_file: Optional[pathlib.Path] = Option(None, '--env-file', help=(
        "Load configuration from a given .env file."
    )),
    version: bool = Option(False, help="Show version number."),
    log_file: Optional[pathlib.Path] = Option(None, '--log-file', help=(
        "Write log messages to a specified file, if not given, writes logs to "
        "STDERR."
    )),
    log_level: Optional[str] = Option('warning', '--log-level', help=(
        "Log level. Possible levels: fatal, error, warning, info, debug. "
        "Default: warning."
    )),
):
    logging.basicConfig(
        level=logging.getLevelName(log_level.upper()),
        format='%(asctime)s %(levelname)s: %(message)s',
        filename=log_file,
    )

    log.debug("log file set to: %s", log_file or 'STDERR')
    log.debug("log level set to: %s", log_level)

    ctx.obj = ctx.obj or create_context('cli', args=option, envfile=env_file)
    if version:
        echo(spinta.__version__)





from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Array
from spinta.backends.postgresql.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL


@commands.wipe.register(Context, Array, PostgreSQL)
def wipe(context: Context, dtype: Array, backend: PostgreSQL):
    wipe(context, dtype.items.dtype, backend)
    table = backend.get_table(dtype.prop, TableType.LIST)
    connection = context.get('transaction').connection
    connection.execute(table.delete())



from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import ShortExpr


def merge_formulas(a: Expr, b: Expr) -> Expr:
    if a and b:
        if a.name == 'and' and b.name == 'and':
            args = a.args + b.args
        elif a.name == 'and':
            args = a.args + (b,)
        elif b.name == 'and':
            args = (a,) + b.args
        else:
            args = a, b
        return ShortExpr('and', *args)
    elif a:
        return a
    elif b:
        return b



from pathlib import Path

import click

from spinta import commands
from spinta.auth import auth_server_keys_exists
from spinta.auth import client_exists
from spinta.auth import create_client_file
from spinta.auth import gen_auth_server_keys
from spinta.components import Config
from spinta.components import Context
from spinta.components import Store
from spinta.core.config import DEFAULT_CONFIG_PATH


def _ensure_config_dir(
    path: Path,
    default_auth_client: str,
    *,
    verbose: bool = True,
):
    """Create default config_path directory if it does not exits"""

    # If a non default config directory is given, it should exist.
    if not path.exists() and path != DEFAULT_CONFIG_PATH:
        raise Exception(f"Config dir {path} does not exist!")

    # Ensure clients directory.
    clients_path = (path / 'clients')
    clients_path.mkdir(parents=True, exist_ok=True)

    # Ensure auth server keys.
    if not auth_server_keys_exists(path):
        if verbose:
            click.echo(f"Initializing auth server keys: {path}")
        gen_auth_server_keys(path)

    # Ensure default client.
    if not client_exists(clients_path, default_auth_client):
        if verbose:
            click.echo(f"Initializing default auth client: {path}")
        create_client_file(
            clients_path,
            default_auth_client,
            secret=None,
            scopes=[
                'spinta_getall',
                'spinta_getone',
                'spinta_search',
                'spinta_changes',
            ],
        )


def load_config(
    context: Context,
    *,
    verbose: bool = True,
    ensure_config_dir: bool = False,
) -> Config:
    config = context.get('config')
    commands.load(context, config)
    if ensure_config_dir:
        _ensure_config_dir(
            config.config_path,
            config.default_auth_client,
            verbose=verbose,
        )
    commands.check(context, config)
    return config


def load_store(
    context: Context,
    *,
    verbose: bool = True,
    ensure_config_dir: bool = False,
) -> Store:
    load_config(
        context,
        verbose=verbose,
        ensure_config_dir=ensure_config_dir,
    )
    store = context.get('store')
    commands.load(context, store)
    return store


def load_manifest(
    context: Context,
    *,
    store: Store = None,
    verbose: bool = True,
    ensure_config_dir: bool = False,
    rename_duplicates: bool = False,
    load_internal: bool = True,
) -> Store:
    if store is None:
        store = load_store(
            context,
            verbose=verbose,
            ensure_config_dir=ensure_config_dir,
        )
    if verbose:
        if store.manifest.path:
            click.echo(
                f"Loading {type(store.manifest).__name__} "
                f"manifest {store.manifest.name} "
                f"({store.manifest.path})..."
            )
        else:
            click.echo(
                f"Loading {type(store.manifest).__name__} "
                f"manifest {store.manifest.name}"
            )
    commands.load(
        context, store.manifest,
        rename_duplicates=rename_duplicates,
        load_internal=load_internal,
    )
    commands.link(context, store.manifest)
    commands.check(context, store.manifest)
    return store


def prepare_manifest(
    context: Context,
    *,
    verbose: bool = True,
    ensure_config_dir: bool = False,
) -> Store:
    store = load_manifest(
        context,
        verbose=verbose,
        ensure_config_dir=ensure_config_dir,
    )
    commands.wait(context, store)
    commands.prepare(context, store.manifest)
    return store



import enum


class Level(enum.IntEnum):
    # Data do not exist.
    absent = 0

    # Data exists in any form, for example txt, pdf, etc...
    available = 1

    # Data is structured, for example xls, xml, etc...
    structured = 2

    # Data provided using an open format, csv, tsv, sql, etc...
    open = 3

    # Individual data objects have unique identifiers.
    identifiable = 4

    # Data is linked with a known vocabulary.
    linked = 5



from typing import Optional, Any, Dict, Iterable, Tuple

import logging
import re


log = logging.getLogger(__name__)


class UnknownValue:

    def __str__(self):
        return '[UNKNOWN]'

    __repr__ = __str__


UNKNOWN_VALUE = UnknownValue()


def resolve_context_vars(schema: Dict[str, str], this: Optional[Any], kwargs: dict):
    """Resolve value from given kwargs and schema."""
    # Extend context by calling get_error_context command on first positional
    # argument of Node or Type type.
    if this:
        from spinta import commands
        schema = {
            **commands.get_error_context(this),
            **schema,
        }
        kwargs = {**kwargs, 'this': this}

    added = set()
    context = {}
    if this:
        context['component'] = type(this).__module__ + '.' + type(this).__name__
    for k, path in schema.items():
        path = path or k
        name, *names = path.split('.')
        if name not in kwargs:
            continue
        added.add(name)
        value = kwargs
        for name in [name] + names:
            if name.endswith('()'):
                name = name[:-2]
                func = True
            else:
                func = False
            if isinstance(value, dict):
                value = value.get(name)
            elif hasattr(value, name):
                value = getattr(value, name)
            else:
                value = UNKNOWN_VALUE
                break
            if func:
                value = value()
        if value is not UNKNOWN_VALUE:
            context[k] = value

    for k in set(kwargs) - added:
        v = kwargs[k]
        if not isinstance(v, (int, float, str)):
            v = str(v)
        context[k] = v

    # Return sorted context.
    names = [
        'component',
        'manifest',
        'schema',
        'backend',
        'dataset',
        'resource',
        'model',
        'entity',
        'property',
        'attribute',
        'type',
    ]
    names += [x for x in schema if x not in names]
    names += [x for x in kwargs if x not in names]

    def sort_key(item: Tuple[str, Any]) -> Tuple[int, str]:
        key = item[0]
        try:
            return names.index(key), key
        except ValueError:
            return len(names), key

    return {k: v for k, v in sorted(context.items(), key=sort_key)}


class BaseError(Exception):
    type: str = None
    status_code: int = 500
    template: str = None
    headers: Dict[str, str] = {}
    context: Dict[str, Any] = {}

    def __init__(self, *args, **kwargs):
        if len(args) == 0:
            this = None
        elif len(args) == 1:
            this = args[0]
        else:
            this = None
            log.error("Only one positional argument is alowed, but %d was given.", len(args), stack_info=True)

        self.type = this.type if this and hasattr(this, 'type') else 'system'

        self.context = resolve_context_vars(self.context, this, kwargs)

        # Remove all unknown values.

        try:
            self.message = _render_template(self)
        except KeyError:
            log.exception("Can't render error message for %s.", self.__class__.__name__)
            self.message = self.template

        super().__init__(
            self.message + '\n' +
            ('  Context:\n' if self.context else '') +
            ''.join(
                f'    {k}: {v}\n'
                for k, v in self.context.items()
            )
        )


def error_response(error: BaseError):
    return {
        'type': error.type,
        'code': type(error).__name__,
        'template': error.template,
        'context': error.context,
        'message': error.message,
    }


def _render_template(error: BaseError):
    if error.type in error.context:
        context = {
            **error.context,
            'this': f'<{error.type} name={error.context[error.type]!r}>',
        }
    else:
        context = error.context
    try:
        return error.template.format(**context)
    except KeyError:
        context = context.copy()
        template_vars_re = re.compile(r'\{(\w+)')
        for match in template_vars_re.finditer(error.template):
            name = match.group(1)
            if name not in context:
                context[name] = UNKNOWN_VALUE
        return error.template.format(**context)




class MultipleErrors(Exception):

    def __init__(self, errors: Iterable[BaseError]):
        self.errors = list(errors)
        super().__init__(
            'Multiple errors:\n' + ''.join([
                f' - {error.message}\n' +
                '     Context:\n' + ''.join(
                    f'       {k}: {v}\n' for k, v in error.context.items()
                )
                for error in self.errors
            ])
        )


class UserError(BaseError):
    status_code = 400


class ConflictingValue(UserError):
    status_code = 409
    template = "Conflicting value."
    context = {
        'given': None,
        'expected': None,
    }


class UniqueConstraint(UserError):
    template = "Given value already exists."


class InvalidOperandValue(UserError):
    template = "Invalid operand value for {operator!r} operator."


class UnknownOperator(UserError):
    template = "Unknown operator {operator!r}."


class ItemDoesNotExist(UserError):
    status_code = 404
    template = "Resource {_id!r} not found."


class ModelNotFound(UserError):
    status_code = 404
    template = "Model {model!r} not found."


class PropertyNotFound(UserError):
    status_code = 404
    template = "Property {property!r} not found."


class NoItemRevision(UserError):
    template = "'_revision' must be given on rewrite operation."


class JSONError(UserError):
    template = "Invalid JSON."


class InvalidValue(UserError):
    template = "Invalid value."


class ValueNotInEnum(UserError):
    template = "Given value {value} is not defined in enum."


class UndefinedEnum(UserError):
    template = "Enum {name!r} is not defined."


class EmptyStringSearch(UserError):
    template = \
        "Empty string can't be used with `{op}`. " \
        "Use `exact` parameter."


class InvalidToken(UserError):
    status_code = 401
    template = "Invalid token"
    headers = {'WWW-Authenticate': 'Bearer error="invalid_token"'}


class BasicAuthRequired(UserError):
    status_code = 401
    template = "Unauthorized"
    headers = {'WWW-Authenticate': 'Basic realm="Authentication required."'}


class AuthorizedClientsOnly(UserError):
    status_code = 401
    template = "This resource can only be accessed by an authorized client."


class MultipleRowsFound(BaseError):
    template = "Multiple rows were found."


class ManagedProperty(UserError):
    template = "Value of this property is managed automatically and cannot be set manually."


class InvalidManagedPropertyName(UserError):
    template = "Invalid managed property name. Expected name {name!r}, got {property!r}."


class NotFoundError(BaseError):
    template = "No results where found."
    status_code = 500


class NodeNotFound(UserError):
    template = "Node {name!r} of type {type!r} not found."


class ModelReferenceNotFound(BaseError):
    template = "Model reference {ref!r} not found."


class ModelReferenceKeyNotFound(BaseError):
    template = "Model reference key {ref!r} not found in {model!r}."


class SourceNotSet(UserError):
    status_code = 404
    template = (
        "Dataset {dataset!r} resource {resource!r} source is not set. "
        "Make sure {resource!r} name parameter in {path} or environment variable {envvar} is set."
    )


class InvalidManifestFile(BaseError):
    template = "Error while parsing {eid!r} manifest entry: {error}"


class ManifestFileDoesNotExist(BaseError):
    template = "Manifest file {path} does not exist."


class UnknownProjectOwner(BaseError):
    template = "Unknown owner {owner}."
    context = {
        'owner': 'this.owner'
    }


class UnknownProjectDataset(BaseError):
    template = "Unknown project dataset."
    context = {
        'manifest': 'project.parent.name',
        'dataset': 'project.dataset',
        'project': 'project.name',
        'filename': 'project.path',
    }


class MissingRequiredProperty(BaseError):
    template = "Property {prop!r} is required."


class FileNotFound(UserError):
    template = "File {file!r} not found."


class UnknownModelReference(BaseError):
    template = "Unknown model reference given in {param!r} parameter."


class InvalidDependencyValue(BaseError):
    template = "Dependency must be in 'object/name.property' form, got: {dependency!r}."


class MultipleModelsInDependencies(BaseError):
    template = "Dependencies are allowed only from single model, but more than one model found: {models}."


class MultipleCommandCallsInDependencies(BaseError):
    template = "Only one command call is allowed."


class MultipleCallsOrModelsInDependencies(BaseError):
    template = "Only one command call or one model is allowed in dependencies."


class InvalidSource(BaseError):
    template = "Invalid source. {error}"
    context = {
        'source': 'source.type',
    }


class UnknownParameter(BaseError):
    template = "Unknown parameter {parameter!r}."
    context = {
        'parameter': 'param',
    }


class InvalidParameterValue(BaseError):
    template = "Invalid parameter {parameter!r} value."


class TooManyParameters(BaseError):
    template = (
        "Too many parameters, you can only give up to {max_params} parameters."
    )


class FieldNotInResource(UserError):
    template = "Unknown property {property!r}."
    context = {
        'property': 'prop',
    }


class UnknownContentType(UserError):
    status_code = 415
    template = "Unknown content type {content_type!r}."


class UnknownAction(UserError):
    template = "Unknown action {action!r}."


class UnknownOutputFormat(UserError):
    template = "Unknown output format {name}."


class UnknownRequestParameter(UserError):
    template = "Unknown request parameter {name!r}."


class OutOfScope(UserError):
    template = "{this} is out of given scope {scope!r}."
    context = {
        'scope': 'scope.name',
    }


class UnhandledException(BaseError):
    status_code = 500
    template = "Unhandled exception {exception}: {error}."
    context = {
        'exception': 'error.__class__.__name__',
    }


class NewItemHasIdAlready(UserError):
    template = "New item has id already set"


class NewItemHasRevisionAlready(UserError):
    template = "New item has revision already set"


class ModelSourceNotFound(UserError):
    template = "Model source {table!r} not found."


class UnavailableSubresource(UserError):
    template = "Subresources only of type Object and File are accessible"


class InvalidPostUrl(UserError):
    template = "Invalid URL for POST request"


class MultipleParentsError(Exception):
    pass


class UnknownPropertyType(UserError):
    template = "Unknown property type {type!r}."


class UnknownExpr(UserError):
    template = "Unknown expression {expr!r}."


class MissingReferrence(UserError):
    template = "Missing reference {ref!r} referenced from {param!r} parameter."


class UnacceptableFileName(UserError):
    template = "Path is not acceptable in filename {file!r}"


class NoAuthServer(UserError):
    template = "Authorization server is disabled, use an external authorization server."


class NoTokenValidationKey(UserError):
    template = "A token validation key is required, set it via token_validation_key configuration parameter."


class NoExternalName(UserError):
    template = "Property {property} does not have 'external' name."


class NoKeyMap(UserError):
    template = "Key map is not configured."


class UnknownKeyMap(UserError):
    template = "Keymap {keymap!r} is not found."


class BackendNotFound(UserError):
    template = "Can't find backend {name!r}."


class NoBackendConfigured(UserError):
    template = "Backend is not configured, can't proceed the request."


class UnexpectedFormulaResult(UserError):
    template = (
        "Unexpected formula {formula} result. "
        "Expected {expected}, instead got a {received}."
    )


class FormulaError(UserError):
    template = "Error while interpreting formula {formula}: {error}."


class UnknownBind(FormulaError):
    template = "Unknown bind {name!r}."


class RequiredConfigParam(UserError):
    template = "Configuration parameter {name!r} is required."


class IncompatibleForeignProperties(UserError):
    template = (
        "Can't join {this} and {right}, these two properties does not have "
        "direct connection with one another."
    )


class KeymapNotSet(UserError):
    template = (
        "Keymap is required for {this}, but is not configured. Please make "
        "sure a key map is configured."
    )


class RemoteClientError(UserError):
    pass


class RemoteClientCredentialsNotFound(RemoteClientError):
    template = (
        "Remote client credentials not found for {name!r}. Make sure, a "
        "section [{section}] exists in {credentials} file. You can use "
        "`spinta remote add {name}` command to add it."
    )


class RemoteClientCredentialsNotGiven(RemoteClientError):
    template = (
        "Make sure client name and secret is given in {credentials} file, "
        "[{section}] section."
    )


class RemoteClientScopesNotGiven(RemoteClientError):
    template = (
        "Make sure at least one scope is given for [{section}] in "
        "{credentials} file."
    )



from __future__ import annotations

from typing import Any, Optional, List

import importlib
import functools
from typing import Dict
from typing import TYPE_CHECKING
from typing import Tuple

from spinta.dispatcher import Command
from spinta import spyna
from spinta.exceptions import UnknownExpr
from spinta.utils.schema import NA

if TYPE_CHECKING:
    from spinta.components import Context


class Expr:
    name: str
    args: Tuple[Any]
    kwargs: Dict[str, Any]

    def __init__(self, *args, **kwargs):
        # Can't use name arguments, because it can be used in kwargs, so
        # expr.name is extracted from args.
        self.name, *args = args
        self.args = tuple(args)
        self.kwargs = kwargs

    def __repr__(self):
        return str(spyna.unparse(self.todict(), raw=True))

    def __str__(self):
        return str(spyna.unparse(self.todict()))

    def todict(self) -> dict:
        args = [
            v.todict() if isinstance(v, Expr) else v
            for v in self.args
        ]
        kwargs = [
            {
                'name': 'bind',
                'args': [k, v.todict() if isinstance(v, Expr) else v],
            } for k, v in self.kwargs.items()
        ]
        return {
            'name': self.name,
            'args': args + kwargs,
        }

    def __call__(self, *args, **kwargs):
        return type(self)(self.name, *args, **kwargs)

    def resolve(self, env: Env):
        args = []
        kwargs = {}
        for arg in self.args:
            arg = env.resolve(arg)
            if isinstance(arg, Pair):
                kwargs[arg.name] = arg.value
            else:
                args.append(arg)
        for key, arg in self.kwargs.items():
            val = env.resolve(arg)
            kwargs[key] = val
        return args, kwargs


class ShortExpr(Expr):
    """Short form of an expression

    For example `a=2` is a short form of `eq(bind('a'),2)`.
    """

    def todict(self) -> dict:
        return {
            **super().todict(),
            'type': 'expression',
        }


class MethodExpr(Expr):
    """Method call expression

    For example `a.b()` instead of `b(a)`.
    """

    def todict(self) -> dict:
        return {
            **super().todict(),
            'type': 'method',
        }


def unparse(expr: Any):
    if expr is NA:
        return None
    if isinstance(expr, Expr):
        expr = expr.todict()
    return spyna.unparse(expr)


class Ufunc(Command):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.autoargs = True


class UFuncRegistry:

    def __init__(self, ufuncs=None):
        self._ufuncs = ufuncs or []

    def __call__(
        self,
        *types,
        name: Optional[str] = None,
        names: Optional[List[str]] = None,
    ):
        def decorator(func):
            if names is None:
                self._ufuncs.append((name or func.__name__, func, types))
            else:
                for name_ in names:
                    func_ = _inject_name(func, name_)
                    self._ufuncs.append((name_, func_, types))
            return func
        return decorator

    def collect(self, modules: List[str]):
        # Collect all ufuncs from given list of modules.
        for module in modules:
            # It is enough to just import module, ufuncs will be collected
            # automatically by __call__.
            importlib.import_module(module)

    def ufuncs(self):
        # Create commands from collected ufuncs.
        ufuncs = {}
        for name, func, types in self._ufuncs:
            if name not in ufuncs:
                ufuncs[name] = Ufunc(name)
            dispatcher = ufuncs[name]
            dispatcher.add(types, func)
            if (
                len(types) > 1 and
                not isinstance(types[1], tuple) and
                issubclass(types[1], Expr)
            ):
                # If function is registered with @resolve(Env, Expr), then
                # disable autoargs. This means, that arguments must be resolved
                # manually by calling ufunc(env, expr).
                dispatcher.autoargs = False
        return ufuncs


def _inject_name(func, name):
    def wrapper(env, *args, **kwargs):
        return func(env, name, *args, **kwargs)
    return wrapper


class ufunc:
    resolver = UFuncRegistry()
    executor = UFuncRegistry()


class Env:
    this: Any
    context: Context

    def __init__(
        self,
        context: Context,
        resolvers: UFuncRegistry = None,
        executors: UFuncRegistry = None,
        scope=None,
    ):
        if resolvers is None or executors is None:
            config = context.get('config')
            resolvers = resolvers or config.resolvers
            executors = executors or config.executors
        self.context = context
        self._resolvers = resolvers
        self._executors = executors
        self._scope = scope or {}

    def __call__(self, **scope):
        return type(self)(
            self.context,
            self._resolvers,
            self._executors,
            {**self._scope, **scope},
        )

    def __getattr__(self, name):
        return self._scope[name]

    def update(self, **scope):
        self._scope.update(scope)

    def default_resolver(self, expr: Expr, *args, **kwargs):
        return expr(*args, **kwargs)

    def call(self, name, *args, **kwargs):
        if name not in self._resolvers:
            raise UnknownExpr(expr=str(Expr(name, *args, **kwargs)), name=name)
        ufunc = self._resolvers[name]
        try:
            return ufunc(self, *args, **kwargs)
        except NotImplementedError:
            raise UnknownExpr(expr=str(Expr(name, *args, **kwargs)), name=name)

    def resolve(self, expr: Any):
        if not isinstance(expr, Expr):
            # Expression is already resolved, return resolved value.
            return expr

        if expr.name in self._resolvers:
            ufunc = self._resolvers[expr.name]

        else:
            args, kwargs = expr.resolve(self)
            return self.default_resolver(expr, *args, **kwargs)

        if ufunc.autoargs:
            # Resolve arguments automatically.
            args, kwargs = expr.resolve(self)
            try:
                return ufunc(self, *args, **kwargs)
            except NotImplementedError:
                return self.default_resolver(expr, *args, **kwargs)

        else:
            # Resolve arguments manually.
            return ufunc(self, expr)

    def execute(self, expr: Any):
        if not isinstance(expr, Expr):
            return expr

        if expr.name not in self._executors:
            raise UnknownExpr(expr=str(expr), name=expr.name)

        ufunc = self._executors[expr.name]

        if ufunc.autoargs:
            args = [self.execute(v) for v in expr.args]
            kwargs = {k: self.execute(v) for k, v in expr.kwargs.items()}
            return ufunc(self, *args, **kwargs)

        else:
            return ufunc(self, expr)


def asttoexpr(ast) -> Expr:
    if isinstance(ast, dict):
        args = [asttoexpr(x) for x in ast['args']]
        typ = ast.get('type')
        if typ == 'expression':
            return ShortExpr(ast['name'], *args)
        elif typ == 'method':
            return MethodExpr(ast['name'], *args)
        else:
            return Expr(ast['name'], *args)
    else:
        return ast


class Bind:

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name


class Pair:

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return f'{self.name}: {self.value!r}'


class Negative(Bind):
    pass


class Positive(Bind):
    pass


bind = functools.partial(Expr, 'bind')



from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Object
from spinta.backends.postgresql.components import PostgreSQL


@commands.prepare.register(Context, PostgreSQL, Object)
def prepare(context: Context, backend: PostgreSQL, dtype: Object):
    columns = []
    for prop in dtype.properties.values():
        if prop.name.startswith('_') and prop.name not in ('_revision',):
            continue
        column = commands.prepare(context, backend, prop)
        if isinstance(column, list):
            columns.extend(column)
        elif column is not None:
            columns.append(column)
    return columns



from spinta import commands
from spinta.components import Context
from spinta.manifests.yaml.components import YamlManifest


@commands.bootstrap.register(Context, YamlManifest)
def bootstrap(context: Context, manifest: YamlManifest):
    # Yaml manifest can't store state so we always run bootstrap.
    store = context.get('store')
    for backend in store.backends.values():
        commands.bootstrap(context, backend)



from typing import Any
from typing import Dict
from urllib.parse import urlencode

from geoalchemy2 import WKBElement
from geoalchemy2.shape import to_shape

from spinta import commands
from spinta.formats.components import Format
from spinta.components import Action
from spinta.components import Context
from spinta.formats.html.components import Cell
from spinta.formats.html.components import Html
from spinta.types.geometry.components import Geometry


@commands.prepare_dtype_for_response.register(Context, Geometry, Format, WKBElement)
def prepare_dtype_for_response(
    context: Context,
    dtype: Geometry,
    fmt: Format,
    value: WKBElement,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    shape = to_shape(value)
    return shape.wkt


@commands.prepare_dtype_for_response.register(Context, Geometry, Html, WKBElement)
def prepare_dtype_for_response(
    context: Context,
    dtype: Geometry,
    fmt: Html,
    value: WKBElement,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    shape = to_shape(value)
    point = shape.centroid
    params = urlencode({'mlat': point.x, 'mlon': point.y})
    link = f'https://www.openstreetmap.org/?{params}#map=19/{point.x}/{point.y}'
    return Cell(shape.wkt, link=link)



from typing import List

from spinta.utils.aiotools import adrain
from spinta import commands
from spinta.components import Context
from spinta.commands.write import push_stream
from spinta.manifests.components import Manifest
from spinta.manifests.backend.helpers import read_sync_versions
from spinta.manifests.backend.helpers import versions_to_dstream
from spinta.manifests.backend.components import BackendManifest


@commands.sync.register()
async def sync(
    context: Context,
    manifest: BackendManifest,
    *,
    sources: List[Manifest] = None,
):
    stream = read_sync_versions(context, manifest)
    stream = versions_to_dstream(manifest, stream)
    await adrain(push_stream(context, stream))



from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name, get_table_name
from spinta.components import Context, Model
from spinta.types.datatype import Ref
from spinta.migrations import SchemaVersion


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, type(None), Ref)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: type(None),
    current: Ref,
):
    return [
        {
            'name': 'column',
            'args': [
                f'{get_column_name(current.prop)}._id',
                {
                    'name': 'ref',
                    'args': [f'{get_table_name(current.model)}._id'],
                },
            ]
        },
    ]


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, Model, Ref)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: Model,
    current: Ref,
):
    column = {
        'name': 'column',
        'args': [
            f'{get_column_name(current.prop)}._id',
            {
                'name': 'ref',
                'args': [f'{get_table_name(current.model)}._id'],
            }
        ]
    }

    upgrade = {
        'name': 'add_column',
        'args': [
            get_table_name(current.prop),
            column,
        ]
    }

    version.actions.append({
        'type': 'schema',
        'upgrade': upgrade,
        'downgrade': {
            'name': 'drop_column',
            'args': [
                get_table_name(current.prop),
                f'{get_column_name(current.prop)}._id',
            ]
        }
    })


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, Ref, Ref)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: Ref,
    current: Ref,
):
    if freezed.model.name != current.model.name:
        raise NotImplementedError("Could not change model's reference")



from __future__ import annotations

import base64
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import TypedDict
from typing import Union

import sqlalchemy as sa
from sqlalchemy.sql.functions import Function

from spinta.auth import authorized
from spinta.components import Action
from spinta.components import Model
from spinta.components import Property
from spinta.core.ufuncs import Bind
from spinta.core.ufuncs import Env
from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import Negative
from spinta.core.ufuncs import ufunc
from spinta.datasets.backends.sql.components import Sql
from spinta.datasets.backends.sql.ufuncs.components import SqlResultBuilder
from spinta.dimensions.enum.helpers import prepare_enum_value
from spinta.dimensions.param.components import ResolvedParams
from spinta.exceptions import PropertyNotFound
from spinta.exceptions import UnknownExpr
from spinta.types.datatype import DataType
from spinta.types.datatype import PrimaryKey
from spinta.types.datatype import Ref
from spinta.types.file.components import FileData
from spinta.ufuncs.components import ForeignProperty
from spinta.utils.data import take
from spinta.utils.itertools import flatten


def ensure_list(value: Any) -> List[Any]:
    if isinstance(value, tuple):
        return list(value)
    if isinstance(value, list):
        return value
    else:
        return [value]


class SqlFrom:
    backend: Sql
    joins: Dict[str, sa.Table]
    from_: sa.Table

    def __init__(self, backend: Sql, table: sa.Table):
        self.backend = backend
        self.joins = {}
        self.from_ = table

    def get_table(
        self,
        env: SqlQueryBuilder,
        prop: ForeignProperty,
    ) -> sa.Table:
        fpr: Optional[ForeignProperty] = None
        for fpr in prop.chain:
            if fpr.name in self.joins:
                continue

            # Left table foreign keys
            lmodel = fpr.left.prop.model
            lenv = env(model=lmodel)
            lfkeys = lenv.call('join_table_on', fpr.left.prop)
            lfkeys = ensure_list(lfkeys)

            # Right table primary keys
            rpkeys = []
            rmodel = fpr.right.prop.model
            renv = env(model=rmodel)
            for rpk in fpr.left.refprops:
                rpkeys += ensure_list(
                    renv.call('join_table_on', rpk)
                )

            # Number of keys on both left and right must be equal.
            assert len(lfkeys) == len(rpkeys), (lfkeys, rpkeys)
            condition = []
            for lfk, rpk in zip(lfkeys, rpkeys):
                condition += [lfk == rpk]

            # Build `JOIN rtable ON (condition)`.
            assert len(condition) > 0
            if len(condition) == 1:
                condition = condition[0]
            else:
                condition = sa.and_(*condition)

            rtable = self.backend.get_table(rmodel)
            self.joins[fpr.name] = self.from_.join(rtable, condition)
            self.from_ = self.joins[fpr.name]

        model = fpr.right.prop.model
        table = self.backend.get_table(model)
        return table


class SqlQueryBuilder(Env):
    backend: Sql
    model: Model
    table: sa.Table
    joins: SqlFrom
    columns: List[sa.Column]
    # `resolved` is used to map which prop.place properties are already
    # resolved, usually it maps to Selected, but different DataType's can return
    # different results.
    resolved: Dict[str, Selected]
    selected: Dict[str, Selected] = None
    params: ResolvedParams

    def init(self, backend: Sql, table: sa.Table):
        return self(
            backend=backend,
            table=table,
            columns=[],
            resolved={},
            selected=None,
            joins=SqlFrom(backend, table),
            sort=[],
            limit=None,
            offset=None,
        )

    def build(self, where):
        if self.selected is None:
            # If select list was not explicitly given by client, then select all
            # properties.
            self.call('select', Expr('select'))

        qry = sa.select(self.columns)
        qry = qry.select_from(self.joins.from_)

        if where is not None:
            qry = qry.where(where)

        if self.sort:
            qry = qry.order_by(*self.sort)

        if self.limit is not None:
            qry = qry.limit(self.limit)

        if self.offset is not None:
            qry = qry.offset(self.offset)

        return qry

    def default_resolver(self, expr, *args, **kwargs):
        raise UnknownExpr(expr=str(expr(*args, **kwargs)), name=expr.name)

    def add_column(self, column: Union[sa.Column, Function]) -> int:
        """Returns position in select column list, which is stored in
        Selected.item.
        """
        assert isinstance(column, (sa.Column, Function)), column
        if column not in self.columns:
            self.columns.append(column)
        return self.columns.index(column)


class Selected:
    # Item index in select list.
    item: int = None
    # Model property if a property is selected.
    prop: Property = None
    # A value or a an Expr for further processing on selected value.
    # TODO: Probably default `prep` value should be `na`.
    prep: Any = None

    def __init__(
        self,
        item: int = None,
        prop: Property = None,
        # `prop` can be Expr or any other value.
        # TODO: Probably default `prep` value should be `na`.
        prep: Any = None,
    ):
        self.item = item
        self.prop = prop
        self.prep = prep

    def __repr__(self):
        return self.debug()

    def debug(self, indent: str = ''):
        prop = self.prop.place if self.prop else 'None'
        if isinstance(self.prep, Selected):
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep=...)\n'
                   ) + self.prep.debug(indent + '  ')
        elif isinstance(self.prep, (tuple, list)):
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep={type(self.prep).__name__}...)\n'
            ) + ''.join([
                p.debug(indent + '- ')
                if isinstance(p, Selected)
                else str(p)
                for p in self.prep
            ])
        else:
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep={self.prep})\n'
            )


@ufunc.resolver(SqlQueryBuilder, Bind, Bind, name='getattr')
def getattr_(env: SqlQueryBuilder, field: Bind, attr: Bind):
    if field.name not in env.model.properties:
        raise PropertyNotFound(env.model, property=field.name)
    prop = env.model.properties[field.name]
    return env.call('getattr', prop.dtype, attr)


@ufunc.resolver(SqlQueryBuilder, Ref, Bind, name='getattr')
def getattr_(env: SqlQueryBuilder, dtype: Ref, attr: Bind) -> ForeignProperty:
    if attr.name not in dtype.model.properties:
        raise PropertyNotFound(dtype.model, property=attr.name)
    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype, prop.dtype)


@ufunc.resolver(SqlQueryBuilder, str, object, names=['eq', 'ne'])
def eq(env: SqlQueryBuilder, op: str, field: str, value: Any):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    return env.call(op, Bind(field), value)


def _sa_compare(op: str, column: sa.Column, value: Any):
    if op == 'eq':
        return column == value

    if op == 'ne':
        return column != value

    if op == 'lt':
        return column < value

    if op == 'le':
        return column <= value

    if op == 'gt':
        return column > value

    if op == 'ge':
        return column >= value

    if op == 'contains':
        return column.contains(value)

    if op == 'startswith':
        return column.startswith(value)

    raise NotImplementedError


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]


T = TypeVar('T')


def _prepare_value(prop: Property, value: T) -> Union[T, List[T]]:
    return prepare_enum_value(prop, value)


@ufunc.resolver(SqlQueryBuilder, Bind, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, field: Bind, value: Any):
    prop = env.model.properties[field.name]
    value = _prepare_value(prop, value)
    return env.call(op, prop.dtype, value)


@ufunc.resolver(SqlQueryBuilder, Bind, list, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, field: Bind, value: List[Any]):
    prop = env.model.properties[field.name]
    value = list(flatten(_prepare_value(prop, v) for v in value))
    return env.call(op, prop.dtype, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, fpr: ForeignProperty, value: Any):
    value = _prepare_value(fpr.right.prop, value)
    return env.call(op, fpr, fpr.right, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, list, names=COMPARE)
def compare(
    env: SqlQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    value: List[Any],
):
    value = list(flatten(_prepare_value(fpr.right.prop, v) for v in value))
    return env.call(op, fpr, fpr.right, value)


@ufunc.resolver(SqlQueryBuilder, DataType, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, dtype: DataType, value: Any):
    column = env.backend.get_column(env.table, dtype.prop)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, object, names=COMPARE)
def compare(
    env: SqlQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: DataType,
    value: Any,
):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, Function, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, func: Function, value: Any):
    return _sa_compare(op, func, value)


@ufunc.resolver(SqlQueryBuilder, DataType, list)
def eq(env: SqlQueryBuilder, dtype: DataType, value: List[Any]):
    column = env.backend.get_column(env.table, dtype.prop)
    return column.in_(value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, list)
def eq(env: SqlQueryBuilder, fpr: ForeignProperty, dtype: DataType, value: list):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop)
    return column.in_(value)


@ufunc.resolver(SqlQueryBuilder, DataType, list)
def ne(env: SqlQueryBuilder, dtype: DataType, value: List[Any]):
    column = env.backend.get_column(env.table, dtype.prop)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, list)
def ne(env: SqlQueryBuilder, fpr: ForeignProperty, value: List[Any]):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, fpr.right.prop)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, Expr, name='and')
def and_(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    if len(args) > 1:
        return sa.and_(*args)
    elif args:
        return args[0]


@ufunc.resolver(SqlQueryBuilder, Expr, name='or')
def or_(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    if len(args) > 1:
        return sa.or_(*args)
    elif args:
        return args[0]


@ufunc.resolver(SqlQueryBuilder, Expr, name='list')
def list_(env: SqlQueryBuilder, expr: Expr) -> List[Any]:
    args, kwargs = expr.resolve(env)
    return list(args)


@ufunc.resolver(SqlQueryBuilder, Expr)
def testlist(env: SqlQueryBuilder, expr: Expr) -> Tuple[Any]:
    args, kwargs = expr.resolve(env)
    return tuple(args)


@ufunc.resolver(SqlQueryBuilder)
def count(env: SqlQueryBuilder):
    return sa.func.count()


@ufunc.resolver(SqlQueryBuilder, Expr)
def select(env: SqlQueryBuilder, expr: Expr):
    keys = [str(k) for k in expr.args]
    args, kwargs = expr.resolve(env)
    args = list(zip(keys, args)) + list(kwargs.items())

    if env.selected is not None:
        raise RuntimeError("`select` was already called.")

    env.selected = {}
    if args:
        for key, arg in args:
            env.selected[key] = env.call('select', arg)
    else:
        for prop in take(['_id', all], env.model.properties).values():
            if authorized(env.context, prop, Action.GETALL):
                env.selected[prop.place] = env.call('select', prop)

    if not env.columns:
        raise RuntimeError(
            f"{expr} didn't added anything to select list."
        )


@ufunc.resolver(SqlQueryBuilder, object)
def select(env: SqlQueryBuilder, value: Any) -> Selected:
    """For things like select(1, count())."""
    return Selected(item=env.add_column(value))


@ufunc.resolver(SqlQueryBuilder, Bind)
def select(env: SqlQueryBuilder, item: Bind, *, nested: bool = False):
    prop = _get_property_for_select(env, item.name, nested=nested)
    return env.call('select', prop)


@ufunc.resolver(SqlQueryBuilder, str)
def select(env: SqlQueryBuilder, item: str, *, nested: bool = False):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    prop = _get_property_for_select(env, item, nested=nested)
    return env.call('select', prop)


def _get_property_for_select(
    env: SqlQueryBuilder,
    name: str,
    *,
    nested: bool = False,
) -> Property:
    # TODO: `name` can refer to (in specified order):
    #       - var - a defined variable
    #       - param - a parameter if parametrization is used
    #       - item - an item of a dict or list
    #       - prop - a property
    #       Currently only `prop` is resolved.
    prop = env.model.flatprops.get(name)
    if prop and (
        # Check authorization only for top level properties in select list.
        # XXX: Not sure if nested is the right property to user, probably better
        #      option is to check if this call comes from a prepare context. But
        #      then how prepare context should be defined? Probably resolvers
        #      should be called with a different env class?
        #      tag:resolving_private_properties_in_prepare_context
        nested or
        authorized(env.context, prop, Action.SEARCH)
    ):
        return prop
    else:
        raise PropertyNotFound(env.model, property=name)


@ufunc.resolver(SqlQueryBuilder, Property)
def select(env: SqlQueryBuilder, prop: Property) -> Selected:
    if prop.place not in env.resolved:
        if isinstance(prop.external, list):
            raise ValueError("Source can't be a list, use prepare instead.")
        if prop.external is None:
            pass
        # TODO: Probably here we should check if `prepare is not NA`. Because
        #       prepare could be set to None by user.
        if prop.external.prepare is not None:
            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = prop.external.prepare
            # XXX: Maybe interpretation of prepare formula should be done under
            #      a different Env? This way, select resolvers would know when
            #      properties are resolved under a formula context and for
            #      example would be able to decide to not check permissions on
            #      properties.
            #      tag:resolving_private_properties_in_prepare_context
            result = env.call('select', prop.dtype, result)
        else:
            # If prepare is not given, then take value from `source`.
            result = env.call('select', prop.dtype)
        assert isinstance(result, Selected), prop
        env.resolved[prop.place] = result
    return env.resolved[prop.place]


@ufunc.resolver(SqlQueryBuilder, DataType)
def select(env: SqlQueryBuilder, dtype: DataType) -> Selected:
    table = env.backend.get_table(env.model)
    column = env.backend.get_column(table, dtype.prop, select=True)
    return Selected(
        item=env.add_column(column),
        prop=dtype.prop,
    )


@ufunc.resolver(SqlQueryBuilder, DataType, object)
def select(env: SqlQueryBuilder, dtype: DataType, prep: Any) -> Selected:
    if isinstance(prep, str):
        # XXX: Backwards compatibility thing.
        #      str values are interpreted as Bind values and Bind values are
        #      assumed to be properties. So here we skip
        #      `env.call('select', prep)` and return `prep` as is.
        #      This should be eventually removed, once backwards compatibility
        #      for resolving strings as properties is removed.
        return Selected(prop=dtype.prop, prep=prep)
    elif isinstance(prep, Expr):
        # If `prepare` expression returns another expression, then this means,
        # it must be processed on values returned by query.
        prop = dtype.prop
        if prop.external and prop.external.name:
            sel = env.call('select', dtype)
            return Selected(item=sel.item, prop=sel.prop, prep=prep)
        else:
            return Selected(item=None, prop=prop, prep=prep)
    else:
        result = env.call('select', prep)
        return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(SqlQueryBuilder, PrimaryKey)
def select(
    env: SqlQueryBuilder,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        # If primary key is not specified use all properties to uniquely
        # identify row.
        pkeys = take(model.properties).values()

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', prop)
    else:
        result = [
            env.call('select', prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(SqlQueryBuilder, Ref, object)
def select(env: SqlQueryBuilder, dtype: Ref, prep: Any) -> Selected:
    fpr = ForeignProperty(None, dtype, dtype.model.properties['_id'].dtype)
    return Selected(
        prop=dtype.prop,
        prep=env.call('select', fpr, fpr.right.prop),
    )


@ufunc.resolver(SqlQueryBuilder, ForeignProperty)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
) -> Selected:
    return env.call('select', fpr, fpr.right.prop)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, Property)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    prop: Property,
) -> Selected:
    resolved_key = fpr / prop
    if resolved_key not in env.resolved:
        if isinstance(prop.external, list):
            raise ValueError("Source can't be a list, use prepare instead.")
        if prop.external.prepare is not None:
            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = prop.external.prepare
            result = env.call('select', fpr, prop.dtype, result)
        else:
            result = env.call('select', fpr, prop.dtype)
        assert isinstance(result, Selected), prop
        env.resolved[resolved_key] = result
    return env.resolved[resolved_key]


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
) -> Selected:
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop, select=True)
    return Selected(
        item=env.add_column(column),
        prop=dtype.prop,
    )


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, object)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    prep: Any,
) -> Selected:
    result = env.call('select', fpr, prep)
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, tuple)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', fpr, v) for v in prep)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, Bind)
def select(env: SqlQueryBuilder, fpr: ForeignProperty, item: Bind):
    model = fpr.right.prop.model
    prop = model.flatprops.get(item.name)
    if prop and authorized(env.context, prop, Action.SEARCH):
        return env.call('select', fpr, prop)
    else:
        raise PropertyNotFound(model, property=item.name)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, PrimaryKey)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        raise RuntimeError(
            f"Can't join {dtype.prop} on right table without primary key."
        )

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', fpr, prop)
    else:
        result = [
            env.call('select', fpr, prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(SqlQueryBuilder, list)
def select(
    env: SqlQueryBuilder,
    prep: List[Any],
) -> List[Any]:
    return [env.call('select', v) for v in prep]


@ufunc.resolver(SqlQueryBuilder, tuple)
def select(
    env: SqlQueryBuilder,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', v) for v in prep)


@ufunc.resolver(SqlQueryBuilder, dict)
def select(
    env: SqlQueryBuilder,
    prep: Dict[str, Any],
) -> Dict[str, Any]:
    # TODO: Add tests.
    return {k: env.call('select', v) for k, v in prep.items()}


@ufunc.resolver(SqlQueryBuilder, Property)
def join_table_on(env: SqlQueryBuilder, prop: Property) -> Any:
    if prop.external.prepare is not None:
        if isinstance(prop.external.prepare, Expr):
            result = env.resolve(prop.external.prepare)
        else:
            result = prop.external.prepare
        return env.call('join_table_on', prop.dtype, result)
    else:
        return env.call('join_table_on', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, DataType)
def join_table_on(env: SqlQueryBuilder, dtype: DataType) -> Tuple[Any]:
    table = env.backend.get_table(dtype.prop.model)
    column = env.backend.get_column(table, dtype.prop)
    return column


@ufunc.resolver(SqlQueryBuilder, PrimaryKey)
def join_table_on(env: SqlQueryBuilder, dtype: DataType) -> Any:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        raise RuntimeError(
            f"Can't join {dtype.prop} on right table without primary key."
        )

    if len(pkeys) == 1:
        prop = pkeys[0]
        return env.call('join_table_on', prop)
    else:
        return [
            env.call('join_table_on', prop)
            for prop in pkeys
        ]


@ufunc.resolver(SqlQueryBuilder, DataType, tuple)
def join_table_on(
    env: SqlQueryBuilder,
    dtype: DataType,
    prep: tuple,
) -> Tuple[Any]:
    return tuple(env.call('join_table_on', v) for v in prep)


@ufunc.resolver(SqlQueryBuilder, Bind)
def join_table_on(env: SqlQueryBuilder, item: Bind):
    prop = env.model.flatprops.get(item.name)
    if not prop or not authorized(env.context, prop, Action.SEARCH):
        raise PropertyNotFound(env.model, property=item.name)
    return env.call('join_table_on', prop)


@ufunc.resolver(SqlQueryBuilder, Bind, name='len')
def len_(env: SqlQueryBuilder, bind: Bind):
    prop = env.model.flatprops[bind.name]
    return env.call('len', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, str, name='len')
def len_(env: SqlQueryBuilder, bind: str):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    prop = env.model.flatprops[bind]
    return env.call('len', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, DataType, name='len')
def len_(env: SqlQueryBuilder, dtype: DataType):
    column = env.backend.get_column(env.table, dtype.prop)
    return sa.func.length(column)


@ufunc.resolver(SqlQueryBuilder, Expr)
def sort(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    env.sort = []
    for key in args:
        prop = env.model.properties[key.name]
        column = env.backend.get_column(env.table, prop)
        if isinstance(key, Negative):
            column = column.desc()
        else:
            column = column.asc()
        env.sort.append(column)


@ufunc.resolver(SqlQueryBuilder, int)
def limit(env: SqlQueryBuilder, n: int):
    env.limit = n


@ufunc.resolver(SqlQueryBuilder, int)
def offset(env: SqlQueryBuilder, n: int):
    env.offset = n


@ufunc.resolver(SqlQueryBuilder, Property, object, object)
def swap(env: SqlQueryBuilder, prop: Property, old: Any, new: Any) -> Any:
    return Expr('swap', old, new)


@ufunc.resolver(SqlQueryBuilder, Expr)
def file(env: SqlQueryBuilder, expr: Expr) -> Expr:
    args, kwargs = expr.resolve(env)
    return Expr(
        'file',
        name=env.call('select', kwargs['name'], nested=True),
        content=env.call('select', kwargs['content'], nested=True),
    )


class _FileSelected(TypedDict):
    name: Selected      # File name
    content: Selected   # File content


@ufunc.resolver(SqlResultBuilder, Expr)
def file(env: SqlResultBuilder, expr: Expr) -> FileData:
    """Post query file data processor

    Will be called with _FileSelected kwargs and no args.
    """
    kwargs: _FileSelected
    args, kwargs = expr.resolve(env)
    assert len(args) == 0, args
    name = env.data[kwargs['name'].item]
    content = env.data[kwargs['content'].item]
    content = base64.b64encode(content).decode()
    return {
        '_id': name,
        # TODO: Content probably should not be returned if not explicitly
        #       requested in select list.
        '_content': content,
    }



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Optional, Tuple, List

import pathlib
import uuid

import jsonpatch

from ruamel.yaml import YAML
from ruamel.yaml.parser import ParserError
from ruamel.yaml.scanner import ScannerError
from ruamel.yaml.error import YAMLError
from ruamel.yaml.scalarstring import walk_tree

from spinta import spyna
from spinta import exceptions
from spinta.exceptions import InvalidManifestFile
from spinta.manifests.yaml.components import InlineManifest
from spinta.utils.itertools import last
from spinta.utils.path import is_ignored
from spinta.components import Context

if TYPE_CHECKING:
    from spinta.migrations import SchemaVersion
    from spinta.manifests.yaml.components import YamlManifest

yaml = YAML(typ='safe')


def yaml_config_params(context: Context, manifest: YamlManifest) -> None:
    # We do not load manifest.path here, because internal manifest loads it
    # differently. Here we only load commong config parameters.
    rc = context.get('rc')
    manifest.ignore = rc.get('ignore', default=[], cast=list)
    manifest.ignore += rc.get('manifests', manifest.name, 'ignore', default=[], cast=list)


def list_yaml_files(manifest: YamlManifest) -> Iterator[pathlib.Path]:
    for file in manifest.path.glob('**/*.yml'):
        if not is_ignored(manifest.ignore, manifest.path, file):
            yield file


def read_yaml_file(path: pathlib.Path):
    try:
        with path.open() as f:
            for data in yaml.load_all(f):
                yield data
    except (ParserError, ScannerError, YAMLError) as e:
        raise exceptions.InvalidManifestFile(
            filename=path,
            error=str(e),
        )


def read_schema_versions(path: pathlib.Path):
    docs = read_yaml_file(path)
    schema = {}
    current = next(docs)
    for version in docs:
        patch = version.get('changes', [])
        patch = jsonpatch.JsonPatch(patch)
        schema = patch.apply(schema)
        if 'id' in current:
            schema['id'] = current['id']
        elif 'id' in schema:
            del schema['id']
        if 'id' not in version:
            raise InvalidManifestFile(
                eid=path,
                error="Version id is not specified.",
            )
        schema['version'] = version['id']
        yield {
            **version,
            'schema': schema,
            'migrate': [
                {
                    **action,
                    'upgrade': spyna.parse(action['upgrade']),
                    'downgrade': spyna.parse(action['downgrade']),
                }
                for action in version['migrate']
            ],
        }


def read_manifest_schemas(
    manifest: YamlManifest,
) -> Iterator[Tuple[pathlib.Path, Optional[dict]]]:
    for path in list_yaml_files(manifest):
        yield path, next(read_yaml_file(path), None)


def read_freezed_manifest_schemas(
    manifest: YamlManifest,
) -> Iterator[Tuple[pathlib.Path, Optional[dict], List[dict]]]:
    for path in list_yaml_files(manifest):
        freezed = last(read_schema_versions(path), None)
        if freezed:
            yield path, freezed['schema']


def read_inline_manifest_schemas(
    manifest: InlineManifest,
) -> Iterator[Tuple[None, Optional[dict]]]:
    for schema in manifest.manifest:
        yield None, schema


def add_new_version(path: pathlib.Path, version: SchemaVersion):

    # Read YAML file
    with path.open() as f:
        versions = yaml.load_all(f)
        versions = list(versions)

    # Update current version
    current = versions[0]

    # Ensure schema id
    if 'id' not in current:
        current['id'] = str(uuid.uuid4())

    # Update current version
    current['version'] = version.id

    # Add new version
    versions.append({
        'id': version.id,
        'date': version.date,
        'parents': version.parents,
        'changes': version.changes,
        'migrate': [
            {
                'type': 'schema',
                'upgrade': spyna.unparse(action['upgrade'], pretty=True),
                'downgrade': spyna.unparse(action['downgrade'], pretty=True),
            }
            for action in version.actions
        ],
    })

    yml = YAML()
    yml.indent(mapping=2, sequence=4, offset=2)
    yml.width = 80
    yml.explicit_start = False

    # Use LiteralScalarString format for all multiline strings.
    walk_tree(versions)

    # Write updates back to YAML file
    with path.open('w') as f:
        yml.dump_all(versions, f)



import cgi

from starlette.requests import Request

from spinta import commands
from spinta.accesslog import AccessLog
from spinta.backends.fs.components import FileSystem
from spinta.backends.mongo.components import WriteTransaction
from spinta.types.file.helpers import prepare_patch_data
from spinta.utils.aiotools import aiter
from spinta.utils.data import take
from spinta.renderer import render
from spinta.components import Action, UrlParams, DataItem
from spinta.types.datatype import DataType, File
from spinta.commands.write import prepare_patch, simple_response, validate_data
from spinta.components import Context, DataSubItem
from spinta.backends.postgresql.files import DatabaseFile
from spinta.backends.postgresql.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL


@commands.push.register(Context, Request, File, PostgreSQL)
async def push(
    context: Context,
    request: Request,
    dtype: File,
    backend: PostgreSQL,
    *,
    action: Action,
    params: UrlParams,
):
    if params.propref:
        return await push[type(context), Request, DataType, type(backend)](
            context, request, dtype, backend,
            action=action,
            params=params,
        )

    prop = dtype.prop

    # XXX: This command should just prepare AsyncIterator[DataItem] and call
    #      push_stream or something like that. Now I think this command does
    #      too much work.

    commands.authorize(context, action, prop)

    transaction: WriteTransaction = context.get('transaction')
    accesslog: AccessLog = context.get('accesslog')
    accesslog.log(
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
        rev=request.headers.get('revision'),
        txn=transaction.id,
    )

    data = DataItem(
        prop.model,
        prop,
        propref=False,
        backend=backend,
        action=action
    )

    if action == Action.DELETE:
        data.given = {
            prop.name: {
                '_id': None,
                '_content_type': None,
                '_content': None,
            }
        }
    else:
        data.given = {
            prop.name: {
                '_content_type': request.headers.get('content-type'),
                '_content': await request.body(),
            }
        }
        if 'Content-Disposition' in request.headers:
            _, cdisp = cgi.parse_header(request.headers['Content-Disposition'])
            if 'filename' in cdisp:
                data.given[prop.name]['_id'] = cdisp['filename']

    if 'Revision' in request.headers:
        data.given['_revision'] = request.headers['Revision']

    commands.simple_data_check(context, data, data.prop, data.model.backend)

    data.saved = commands.getone(context, prop, dtype, prop.model.backend, id_=params.pk)

    dstream = aiter([data])
    dstream = validate_data(context, dstream)
    dstream = prepare_patch(context, dstream)
    if action in (Action.UPDATE, Action.PATCH, Action.DELETE):
        dstream = commands.update(context, prop, dtype, prop.model.backend, dstream=dstream)
        dstream = commands.create_changelog_entry(
            context, prop.model, prop.model.backend, dstream=dstream,
        )
    # XXX: This will never be executed, due to previous conditional.
    #      How it should be done correctly?
    elif action == Action.DELETE:
        dstream = commands.delete(context, prop, dtype, prop.model.backend, dstream=dstream)
        dstream = commands.create_changelog_entry(
            context, prop.model, prop.model.backend, dstream=dstream,
        )

    else:
        raise Exception(f"Unknown action {action!r}.")

    status_code, response = await simple_response(context, params.fmt, dstream)
    return render(context, request, prop, params, response, status_code=status_code)


@commands.before_write.register(Context, File, PostgreSQL)
def before_write(
    context: Context,
    dtype: File,
    backend: PostgreSQL,
    *,
    data: DataSubItem,
):
    content = take('_content', data.patch)
    if isinstance(content, bytes) and isinstance(dtype.backend, PostgreSQL):
        transaction = context.get('transaction')
        connection = transaction.connection
        prop = dtype.prop
        table = backend.get_table(prop, TableType.FILE)
        with DatabaseFile(connection, table, mode='w') as f:
            f.write(data.patch['_content'])
            data.patch['_size'] = f.size
            data.patch['_blocks'] = f.blocks
            data.patch['_bsize'] = f.bsize
    elif isinstance(content, bytes) and isinstance(dtype.backend, FileSystem):
        filepath = dtype.backend.path / data.given['_id']
        with open(filepath, 'wb') as f:
            f.write(content)

    patch = prepare_patch_data(dtype, data)

    return {
        f'{dtype.prop.place}.{k}': v for k, v in patch.items()
    }



from spinta import commands
from spinta.utils.schema import NA
from spinta.utils.data import take
from spinta.types.datatype import Array
from spinta.components import Context, Action, DataSubItem
from spinta.backends.mongo.components import Mongo


@commands.before_write.register(Context, Array, Mongo)
def before_write(
    context: Context,
    dtype: Array,
    backend: Mongo,
    *,
    data: DataSubItem,
):
    if data.root.action == Action.INSERT or (data.root.action == Action.UPSERT and data.saved is NA):
        return {dtype.prop.place: data.patch}
    else:
        if dtype.prop.list:
            return {}
        else:
            return take({dtype.prop.place: data.patch})



import enum


class Access(enum.IntEnum):
    # Private properties can be accesses only if an explicit property scope is
    # given. I client does not have required scope, then private properties
    # can't bet selected, but can be used in query conditions, in sorting.
    private = 0

    # Property is exposed only to authorized user, who has access to model.
    # Authorization token is given manually to each user.
    protected = 1

    # Property can be accessed by anyone, but only after accepting terms and
    # conditions, that means, authorization is still needed and data can only be
    # used as specified in provided terms and conditions. Authorization token
    # can be obtained via WebUI.
    public = 2

    # Open data, anyone can access this data, no authorization is required.
    open = 3



from __future__ import annotations

from typing import TYPE_CHECKING

import sqlalchemy as sa

from spinta.components import Context

if TYPE_CHECKING:
    from spinta.backends.postgresql.components import PostgreSQL


def read_manifest_schemas(context: Context, backend: PostgreSQL, conn):
    meta = sa.MetaData(backend.engine)
    table = sa.Table('_schema', meta, autoload_with=backend.engine)
    query = sa.select([table.c._id, table.c.schema])
    i = 0
    for i, row in enumerate(conn.execute(query), 1):
        yield (
            row[table.c._id],
            row[table.c.schema],
        )
    if i == 0:
        raise Exception(
            "Empty _schema table, nothing is loaded into manifest."
        )



from typing import TypedDict


class FileData(TypedDict, total=False):
    _id: str  # File name
    _content_type: str  # Media type (or MIME-type)
    _content: str  # base64 encoded string of file content



from spinta import spyna
from spinta.core.ufuncs import Env, asttoexpr, ufunc


class UFuncTester:

    def __init__(
        self,
        Env=Env,
        context=None,
        resolver=ufunc.resolver,
        executor=ufunc.executor,
        scope=None,
    ):
        self.Env = Env
        self.context = context
        self.resolver = resolver
        self.executor = executor
        self.scope = scope or {}

    def __call__(self, expr, **scope):
        env = self.Env(
            self.context,
            self.resolver.ufuncs(),
            self.executor.ufuncs(),
            self.scope,
        )
        env.update(**scope)
        ast = spyna.parse(expr)
        expr = asttoexpr(ast)
        expr = env.resolve(expr)
        return env.execute(expr)



from typing import List

from requests.models import Response


def parse_csv(resp: Response) -> List[List[str]]:
    resp.raise_for_status()
    return [
        line.strip().split(',')
        for line in resp.text.splitlines()
    ]



import collections
from typing import NamedTuple
from typing import Union

import itertools
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import TypedDict

from starlette.requests import Request
from starlette.responses import Response
from toposort import toposort

from spinta import commands
from spinta import exceptions
from spinta.auth import authorized
from spinta.backends.helpers import get_select_prop_names
from spinta.backends.helpers import get_select_tree
from spinta.backends.components import BackendFeatures
from spinta.compat import urlparams_to_expr
from spinta.components import Action
from spinta.components import Config
from spinta.components import Context
from spinta.components import Model
from spinta.components import Namespace
from spinta.components import Node
from spinta.components import UrlParams
from spinta.manifests.components import Manifest
from spinta.nodes import load_node
from spinta.renderer import render
from spinta.types.datatype import Ref


class NamespaceData(TypedDict):
    title: str
    description: str


def load_namespace_from_name(
    context: Context,
    manifest: Manifest,
    path: str,
    *,
    # Drop last element from path which is usually a model name.
    drop: bool = True,
) -> Namespace:
    ns: Optional[Namespace] = None
    parent: Optional[Namespace] = None
    parts: List[str] = []
    parts_ = path.split('/')
    if drop:
        parts_ = parts_[:-1]
    for part in [''] + parts_:
        parts.append(part)
        name = '/'.join(parts[1:])
        if name not in manifest.namespaces:
            ns = Namespace()
            data = {
                'type': 'ns',
                'name': name,
                'title': '',
                'description': '',
            }
            commands.load(context, ns, data, manifest)
            ns.generated = True
        else:
            ns = manifest.namespaces[name]
            pass

        ns.parent = parent or manifest

        if parent and part and part not in parent.names:
            parent.names[part] = ns

        parent = ns

    return ns


@commands.load.register(Context, Namespace, dict, Manifest)
def load(
    context: Context,
    ns: Namespace,
    data: Dict[str, Any],
    manifest: Manifest,
    *,
    source: Manifest = None,
):
    load_node(context, ns, data)
    ns.path = manifest.path
    ns.manifest = manifest
    ns.access = manifest.access
    ns.backend = None
    ns.names = {}
    ns.models = {}
    manifest.namespaces[ns.name] = ns


@commands.link.register(Context, Namespace)
def link(context: Context, ns: Namespace):
    pass


@commands.check.register(Context, Namespace)
def check(context: Context, ns: Namespace):
    pass


@commands.authorize.register(Context, Action, Namespace)
def authorize(context: Context, action: Action, ns: Namespace):
    authorized(context, ns, action, throw=True)


@commands.getall.register(Context, Namespace, Request)
async def getall(
    context: Context,
    ns: Namespace,
    request: Request,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    config: Config = context.get('config')
    if config.root and ns.is_root():
        ns = ns.manifest.namespaces[config.root]

    commands.authorize(context, action, ns)

    if params.all and params.ns:
        accesslog = context.get('accesslog')
        accesslog.log(
            ns=ns.name,
            action=action.value,
        )

        for model in traverse_ns_models(context, ns, action, internal=True):
            commands.authorize(context, action, model)
        return _get_ns_content(
            context,
            request,
            ns,
            params,
            action,
            recursive=True,
        )
    elif params.all:
        accesslog = context.get('accesslog')
        accesslog.log(
            ns=ns.name,
            action=action.value,
        )

        prepare_data_for_response_kwargs = {}
        for model in traverse_ns_models(context, ns, action, internal=True):
            commands.authorize(context, action, model)
            select_tree = get_select_tree(context, action, params.select)
            prop_names = get_select_prop_names(
                context, model, action, select_tree,
            )
            prepare_data_for_response_kwargs[model.model_type()] = {
                'select': select_tree,
                'prop_names': prop_names,
            }
        expr = urlparams_to_expr(params)
        rows = getall(context, ns, action=action, query=expr)
        rows = (
            commands.prepare_data_for_response(
                context,
                ns.manifest.models[row['_type']],
                params.fmt,
                row,
                action=action,
                **prepare_data_for_response_kwargs[row['_type']],
            )
            for row in rows
        )
        return render(context, request, ns, params, rows, action=action)
    else:
        return _get_ns_content(context, request, ns, params, action)


@commands.getall.register(Context, Namespace)
def getall(
    context: Context,
    ns: Namespace,
    *,
    action: Optional[Action] = None,
    dataset_: Optional[str] = None,
    resource: Optional[str] = None,
    **kwargs
):
    return _query_data(context, ns, action, dataset_, resource, **kwargs)


def _query_data(
    context: Context,
    ns: Namespace,
    action: Action,
    dataset_: Optional[str] = None,
    resource: Optional[str] = None,
    **kwargs,
):
    models = traverse_ns_models(
        context,
        ns,
        action,
        dataset_,
        resource,
        internal=True,
    )
    for model in models:
        yield from commands.getall(context, model, model.backend, **kwargs)


def traverse_ns_models(
    context: Context,
    ns: Namespace,
    action: Action,
    dataset_: Optional[str] = None,
    resource: Optional[str] = None,
    *,
    internal: bool = False,
):
    models = (ns.models or {})
    for model in models.values():
        if _model_matches_params(context, model, action, dataset_, resource, internal):
            yield model
    for ns_ in ns.names.values():
        if not internal and ns_.name.startswith('_'):
            continue
        yield from traverse_ns_models(
            context,
            ns_,
            action,
            dataset_,
            resource,
            internal=internal,
        )


def _model_matches_params(
    context: Context,
    model: Model,
    action: Action,
    dataset_: Optional[str] = None,
    resource: Optional[str] = None,
    internal: bool = False,
):
    if not internal and model.name.startswith('_'):
        return False

    if not authorized(context, model, action):
        return False

    if (
        dataset_ is None and
        resource is None
    ):
        return True

    if model.external is None:
        return False

    if (
        dataset_ is not None and (
            model.external.dataset is None or
            model.external.dataset.name != dataset_
        )
    ):
        return False

    if (
        resource is not None and (
            model.external.resource is None or
            model.external.resource.name != resource
        )
    ):
        return False

    return True


def _get_ns_content(
    context: Context,
    request: Request,
    ns: Namespace,
    params: UrlParams,
    action: Action,
    *,
    recursive: bool = False,
    dataset_: Optional[str] = None,
    resource: Optional[str] = None,
) -> Response:
    if recursive:
        data = _get_ns_content_data_recursive(context, ns, action, dataset_, resource)
    else:
        data = _get_ns_content_data(context, ns, action, dataset_, resource)

    data = sorted(data, key=lambda x: (x.data['_type'] != 'ns', x.data['name']))

    model = ns.manifest.models['_ns']
    select = params.select or ['name', 'title', 'description']
    select_tree = get_select_tree(context, action, select)
    prop_names = get_select_prop_names(
        context, model, action, select_tree,
        auth=False,
    )
    data = (
        commands.prepare_data_for_response(
            context,
            model,
            params.fmt,
            row.data,
            action=action,
            select=select_tree,
            prop_names=prop_names,
        )
        for row in data
    )

    return render(context, request, model, params, data, action=action)


class _NodeAndData(NamedTuple):
    node: Union[Namespace, Model]
    data: Dict[str, Any]


def _get_ns_content_data_recursive(
    context: Context,
    ns: Namespace,
    action: Action,
    dataset_: Optional[str] = None,
    resource: Optional[str] = None,
) -> Iterable[_NodeAndData]:
    yield from _get_ns_content_data(context, ns, action, dataset_, resource)
    for name in ns.names.values():
        yield from _get_ns_content_data_recursive(context, name, action, dataset_, resource)


def _get_ns_content_data(
    context: Context,
    ns: Namespace,
    action: Action,
    dataset_: Optional[str] = None,
    resource: Optional[str] = None,
) -> Iterable[_NodeAndData]:
    items: Iterable[Union[Namespace, Model]] = itertools.chain(
        ns.names.values(),
        ns.models.values(),
    )

    for item in items:
        if _model_matches_params(context, item, action, dataset_, resource):
            yield _NodeAndData(item, {
                '_type': item.node_type(),
                'name': item.model_type(),
                'title': item.title,
                'description': item.description,
            })


@commands.getone.register(Context, Request, Namespace)
async def getone(
    context: Context,
    request: Request,
    ns: Namespace,
    *,
    action: Action,
    params: UrlParams,
):
    raise exceptions.ModelNotFound(model=ns.name)


@commands.in_namespace.register()  # noqa
def in_namespace(node: Node, ns: Namespace) -> bool:
    return node.name.startswith(ns.name)


@commands.in_namespace.register()  # noqa
def in_namespace(node: Node, parent: Node) -> bool:  # noqa
    while node:
        if node is parent:
            return True
        node = node.parent
    return False


@commands.wipe.register(Context, Namespace, type(None))
def wipe(context: Context, ns: Namespace, backend: type(None)):
    commands.authorize(context, Action.WIPE, ns)
    models = traverse_ns_models(context, ns, Action.WIPE, internal=True)
    models = sort_models_by_refs(models)
    for model in models:
        if BackendFeatures.WRITE in model.backend.features:
            commands.wipe(context, model, model.backend)


def sort_models_by_refs(models: Iterable[Model]) -> Iterator[Model]:
    models = {model.model_type(): model for model in models}
    graph = collections.defaultdict(set)
    for name, model in models.items():
        graph[''].add(name)
        for dtype in iter_model_refs(model):
            ref = dtype.model.model_type()
            if ref in models:
                graph[ref].add(name)
    graph = toposort(graph)
    seen = {''}
    for group in graph:
        for name in sorted(group):
            if name in seen:
                continue
            seen.add(name)
            yield models[name]


def iter_model_refs(model: Model) -> Iterator[Ref]:
    for prop in model.properties.values():
        if prop.dtype.name == 'ref':
            yield prop.dtype



from typing import Union

import hashlib

from spinta.components import Model, Property
from spinta.backends.postgresql.constants import TableType
from spinta.backends.postgresql.constants import NAMEDATALEN


def get_table_name(
    node: Union[Model, Property],
    ttype: TableType = TableType.MAIN,
) -> str:
    if isinstance(node, Model):
        model = node
    else:
        model = node.model
    if ttype in (TableType.LIST, TableType.FILE):
        name = model.model_type() + ttype.value + '/' + node.place
    else:
        name = model.model_type() + ttype.value
    return name


def get_column_name(prop: Property) -> str:
    if prop.list:
        if prop.place == prop.list.place:
            return prop.list.name
        else:
            return prop.place[len(prop.list.place) + 1:]
    else:
        return prop.place


def get_pg_name(name: str) -> str:
    if len(name) > NAMEDATALEN:
        hs = 8
        h = hashlib.sha1(name.encode()).hexdigest()[:hs]
        i = int(NAMEDATALEN / 100 * 60)
        j = NAMEDATALEN - i - hs - 2
        name = name[:i] + '_' + h + '_' + name[-j:]
    return name



from texttable import Texttable
from texttable import Texttable


def draw(buffer, name, tnum, width):
    if tnum > 1:
        if name:
            yield f"\n\nTable {name} #{tnum}:\n"
        else:
            yield f"\n\nTable #{tnum}:\n"

    table = Texttable(width)
    table.set_deco(Texttable.HEADER)
    table.add_rows(buffer)
    yield table.draw()



from spinta.commands import load, prepare, check
from spinta.components import Context, Node, MetaData
from spinta.manifests.components import Manifest
from spinta.nodes import load_node
from spinta import exceptions


class Project(MetaData):
    schema = {
        'version': {'type': 'integer', 'required': True},
        'date': {'type': 'date', 'required': True},
        'objects': {'type': 'object', 'default': {}},
        'impact': {'type': 'array', 'default': []},
        'url': {'type': 'url'},
        'source_code': {'type': 'url'},
        'website': {'type': 'url'},
        'owner': {'type': 'string'},
        'dataset': {'type': 'string'},
    }


class Impact:
    schema = {
        'year': {'type': 'integer', 'required': True},
        'users': {'type': 'integer'},
        'revenue': {'type': 'number'},
        'employees': {'type': 'integer'},
        'parent': {'type': 'project'},
    }


class Model(Node):
    schema = {
        'properties': {'type': 'object', 'default': {}},
        'dataset': {'type': 'string', 'inherit': True},
        'target': {'type': 'string'},
    }


class Property(Node):
    schema = {
        'enum': {'type': 'array'},
        'dataset': {'type': 'string', 'inherit': True},
        'target': {'type': 'string'},
    }
    # TODO: inherit type from model if not provided, type is needed for data
    #       serialization.


@load.register(Context, Project, dict, Manifest)
def load(
    context: Context,
    project: Project,
    data: dict,
    manifest: Manifest,
    *,
    source: Manifest = None,
):
    load_node(context, project, data)
    project.impact = [
        {
            'year': None,
            'users': 0,
            'revenue': 0,
            'employees': 0,
            **impact,
        } for i, impact in enumerate(data.get('impact', []))
    ]


@prepare.register(Context, Project)
def prepare(context: Context, project: Project):
    for model in project.objects.values():
        prepare(context, model)


@check.register(Context, Project)
def check(context: Context, project: Project):
    if project.owner and project.owner not in project.manifest.objects['owner']:
        raise exceptions.UnknownOwner(project)

    if project.dataset and project.dataset not in project.manifest.objects['dataset']:
        # TODO add  similar 'dataset' checks for model and property.
        raise exceptions.UnknownProjectDataset(project)



import operator

import itertools

from spinta.formats.components import Format
from spinta.formats.ascii.helpers import draw
from spinta.utils.nestedstruct import flatten


class Ascii(Format):
    content_type = 'text/plain'
    accept_types = {
        'text/plain',
    }
    params = {
        'width': {'type': 'integer'},
        'colwidth': {'type': 'integer'},
    }

    def __call__(self, data, width=None, colwidth=42):
        data = iter(data)
        peek = next(data, None)

        if peek is None:
            return

        data = itertools.chain([peek], data)

        if '_type' in peek:
            groups = itertools.groupby(data, operator.itemgetter('_type'))
        else:
            groups = [(None, data)]

        for name, group in groups:
            if name:
                yield f'\n\nTable: {name}\n'

            rows = flatten(group)
            peek = next(rows, None)
            rows = itertools.chain([peek], rows)

            if name:
                cols = [k for k in peek.keys() if k != '_type']
            else:
                cols = list(peek.keys())

            if colwidth:
                width = len(cols) * colwidth

            buffer = [cols]
            tnum = 1

            for row in rows:
                buffer.append([row.get(c) for c in cols])
                if len(buffer) > 100:
                    yield from draw(buffer, name, tnum, width)
                    buffer = [cols]
                    tnum += 1

            if buffer:
                yield from draw(buffer, name, tnum, width)



from typing import AsyncIterable
from typing import AsyncIterator

import io
import codecs


async def splitlines(
    stream: AsyncIterable[bytes],
    encoding: str = 'utf-8',
    errors: str = 'strict',
) -> AsyncIterator[str]:
    """Read chunks of bytes from stream and yield lines."""
    decoder = codecs.getincrementaldecoder(encoding)(errors)
    decoder = io.IncrementalNewlineDecoder(decoder, translate=True, errors=errors)
    buffer = ''
    async for chunk in stream:
        buffer = buffer + decoder.decode(chunk)
        if '\n' in buffer:
            lines = buffer.split('\n')
            for line in lines[:-1]:
                yield line
            buffer = lines[-1]
    buffer = buffer + decoder.decode(b'', final=True)
    if buffer:
        for line in buffer.split('\n'):
            yield line



from __future__ import annotations

from enum import Enum
from typing import Any
from typing import Dict
from typing import Final
from typing import IO
from typing import List
from typing import Literal
from typing import Optional
from typing import TypedDict

from spinta.dimensions.lang.components import LangData
from spinta.manifests.components import Manifest


class TabularFormat(Enum):
    CSV = 'csv'
    ASCII = 'ascii'
    XLSX = 'xlsx'
    GSHEETS = 'gsheets'


class TabularManifest(Manifest):
    format: TabularFormat = None
    path: str = None


class CsvManifest(TabularManifest):
    type = 'csv'
    format: TabularFormat = TabularFormat.CSV
    file: IO[str] = None

    @staticmethod
    def detect_from_path(path: str) -> bool:
        return path.endswith('.csv')


class AsciiManifest(TabularManifest):
    type = 'ascii'
    format: TabularFormat = TabularFormat.ASCII
    file: IO[str] = None

    @staticmethod
    def detect_from_path(path: str) -> bool:
        return path.endswith('.txt')


class XlsxManifest(TabularManifest):
    type = 'xlsx'
    format: TabularFormat = TabularFormat.XLSX

    @staticmethod
    def detect_from_path(path: str) -> bool:
        return path.endswith('.xlsx')


class GsheetsManifest(TabularManifest):
    type = 'gsheets'
    format: TabularFormat = TabularFormat.GSHEETS

    @staticmethod
    def detect_from_path(path: str) -> bool:
        return path.startswith('https://docs.google.com/spreadsheets/')


ID: Final = 'id'
DATASET: Final = 'dataset'
RESOURCE: Final = 'resource'
BASE: Final = 'base'
MODEL: Final = 'model'
PROPERTY: Final = 'property'
TYPE: Final = 'type'
REF: Final = 'ref'
SOURCE: Final = 'source'
PREPARE: Final = 'prepare'
LEVEL: Final = 'level'
ACCESS: Final = 'access'
URI: Final = 'uri'
TITLE: Final = 'title'
DESCRIPTION: Final = 'description'
ManifestColumn = Literal[
    'id',
    'dataset',
    'resource',
    'base',
    'model',
    'property',
    'type',
    'ref',
    'source',
    'prepare',
    'level',
    'access',
    'uri',
    'title',
    'description',
]
MANIFEST_COLUMNS: List[ManifestColumn] = [
    ID,
    DATASET,
    RESOURCE,
    BASE,
    MODEL,
    PROPERTY,
    TYPE,
    REF,
    SOURCE,
    PREPARE,
    LEVEL,
    ACCESS,
    URI,
    TITLE,
    DESCRIPTION,
]

ManifestRow = Dict[ManifestColumn, str]


class ManifestTableRow(TypedDict, total=False):
    type: str
    backends: Dict[str, BackendRow]


class DatasetRow(TypedDict, total=False):
    type: str
    id: str
    name: str
    level: str
    access: str
    title: str
    description: str
    resources: Dict[str, ResourceRow]
    lang: LangData


class ResourceRow(ManifestRow):
    backend: str
    external: str
    lang: LangData


class BackendRow(TypedDict, total=False):
    type: str
    name: str
    dsn: str
    title: str
    description: str


class BaseRow(TypedDict, total=False):
    model: str
    pk: str
    lang: LangData


class ParamRow(TypedDict):
    name: str                   # param name
    source: List[str]           # list of `self` for prepare formulas
    prepare: List[Any]          # list of formulas
    title: str
    description: str


class ModelExtraData(TypedDict):
    params: List[ParamRow]


class ModelRow(TypedDict, total=False):
    type: str
    id: str
    name: str
    base: Optional[str]
    level: str
    access: str
    title: str
    description: str
    properties: Dict[str, PropertyRow]
    external: ModelExternalRow
    backend: str
    lang: LangData
    data: ModelExtraData


class ModelExternalRow(TypedDict, total=False):
    dataset: str
    resource: str
    pk: List[str]
    name: str
    prepare: Dict[str, Any]


class EnumRow(TypedDict, total=False):
    name: str
    source: str
    prepare: Optional[Dict[str, Any]]
    access: str
    title: str
    description: str
    lang: LangData


class PropertyRow(TypedDict, total=False):
    type: str
    type_args: List[str]
    prepare: Optional[Dict[str, Any]]
    level: str
    access: str
    uri: str
    title: str
    description: str
    model: str
    refprops: List[str]
    external: PropertyExternalRow
    enum: str
    enums: Dict[str, Dict[str, EnumRow]]
    lang: LangData
    units: str


class PropertyExternalRow(TypedDict, total=False):
    name: str
    prepare: Optional[Dict[str, Any]]


class PrefixRow(TypedDict, total=False):
    id: str
    eid: str
    type: str
    name: str
    uri: str
    title: str
    description: str



from typing import Any
from typing import Dict
from typing import Union

from spinta import commands
from spinta.components import Context
from spinta.datasets.components import Dataset
from spinta.dimensions.prefix.components import UriPrefix
from spinta.manifests.components import Manifest
from spinta.nodes import get_node


def load_prefixes(
    context: Context,
    manifest: Manifest,
    node: Union[Manifest, Dataset],
    prefixes: Dict[str, Dict[str, Any]],
) -> Dict[str, UriPrefix]:
    config = context.get('config')
    loaded = {}
    for name, data in prefixes.items():
        prefix: UriPrefix = get_node(
            config,
            manifest,
            data['eid'],
            data,
            group='dimensions',
            parent=node,
        )
        prefix.eid = data['eid']
        prefix.type = data['type']
        prefix.parent = node
        commands.load(context, prefix, data, parent=node)
        loaded[name] = prefix
    return loaded





from spinta import commands
from spinta.components import Context
from spinta.backends.memory.components import Memory


@commands.bootstrap.register(Context, Memory)
def bootstrap(context: Context, backend: Memory):
    pass



from spinta.backends.components import Backend
from spinta.commands import is_object_id, load, load_search_params, load_operator_value, prepare
from spinta.components import Context
from spinta.exceptions import InvalidValue, EmptyStringSearch
from spinta.types.datatype import Array, Integer, DataType, PrimaryKey, String


@load_search_params.register()
def load_search_params(
    context: Context,
    dtype: DataType,
    backend: Backend,
    query_params: dict,
) -> object:
    value = query_params['args'][1]
    value = load(context, dtype, value)
    load_operator_value(context, backend, dtype, value, query_params=query_params)
    return prepare(context, dtype, backend, value)


@load_search_params.register()
def load_search_params(
    context: Context,
    dtype: String,
    backend: Backend,
    query_params: dict,
) -> object:
    value = query_params['args'][1]
    value = load(context, dtype, value)
    operator = query_params['name']
    if operator in ('startswith', 'contains') and value == '':
        raise EmptyStringSearch(dtype, op=operator)
    load_operator_value(context, backend, dtype, value, query_params=query_params)
    return prepare(context, dtype, backend, value)


@load_search_params.register()
def load_search_params(
    context: Context,
    dtype: PrimaryKey,
    backend: Backend,
    query_params: dict,
) -> object:
    value = query_params['args'][1]
    operator = query_params['name']
    if operator in ('startswith', 'contains') and value == '':
        raise EmptyStringSearch(dtype, op=operator)

    if (
        operator not in ('startswith', 'contains') and
        not is_object_id(context, backend, dtype.prop.model, value) or
        value is None
    ):
        raise InvalidValue(dtype)
    load_operator_value(context, backend, dtype, value, query_params=query_params)
    return prepare(context, dtype, backend, value)


@load_search_params.register()
def load_search_params(
    context: Context,
    dtype: Integer,
    backend: Backend,
    query_params: dict,
) -> object:

    # try to convert string search parameter to integer
    # query_params are always string, thus for type loading and validation
    # to work, we need to convert that query value string into appropriate type
    try:
        value = int(query_params['args'][1])
    except ValueError:
        raise InvalidValue(dtype)

    value = load(context, dtype, value)
    load_operator_value(context, backend, dtype, value, query_params=query_params)
    return prepare(context, dtype, backend, value)


@load_search_params.register()
def load_search_params(
    context: Context,
    dtype: Array,
    backend: Backend,
    query_params: dict,
) -> object:
    # if dtype is Array - then we need to work with Array's items
    # i.e. we search a string inside an array, thus we must load string
    # dtype and not the array type.
    value = query_params['args'][1]
    value = load(context, dtype.items.dtype, value)
    load_operator_value(context, backend, dtype.items.dtype, value, query_params=query_params)
    return prepare(context, dtype, backend, value)



import enum
import contextlib
from typing import Any
from typing import Dict
from typing import Optional
from typing import Set


class BackendOrigin(enum.Enum):
    """Origin where backend was defined.

    Backend can be defined in multiple places, for example backend can be
    defined in a configuration file or inline in manifest.
    """

    config = 'config'
    manifest = 'manifest'
    resource = 'resource'


class BackendFeatures(enum.Enum):
    # Files are stored in blocks and file metadata must include _bsize and
    # _blocks properties.
    FILE_BLOCKS = 'FILE_BLOCKS'

    # Backend supports write operations.
    WRITE = 'WRITE'


class Backend:
    metadata = {
        'name': 'backend',
    }

    type: str
    name: str
    origin: BackendOrigin
    features: Set[BackendFeatures] = set()

    # Original configuration values given in manifest, this is used to restore
    # manifest back to its original form.
    config: Dict[str, Any]

    def __repr__(self):
        return (
            f'<{self.__class__.__module__}.{self.__class__.__name__}'
            f'(name={self.name!r}) at 0x{id(self):02x}>'
        )

    @contextlib.contextmanager
    def transaction(self):
        raise NotImplementedError

    def bootstrapped(self):
        raise NotImplementedError


SelectTree = Optional[Dict[str, 'SelectTree']]



import sqlalchemy as sa

from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.dialects.postgresql import ARRAY

from spinta.commands import get_primary_key_type
from spinta.exceptions import UnknownExpr
from spinta.core.ufuncs import Env, Expr, Bind, ufunc


TYPES = {
    'array': ARRAY,
    'string': sa.Text,
    'integer': sa.Integer,
    'number': sa.Float,
    'boolean': sa.Boolean,
    'binary': sa.LargeBinary,
    'date': sa.Date,
    'datetime': sa.DateTime,
    'json': JSONB,
    'uuid': UUID,
}


class Alembic(Env):

    def default_resolver(self, expr, *args, **kwargs):
        raise UnknownExpr(expr=str(expr(*args, **kwargs)), name=expr.name)


@ufunc.resolver(Alembic, Expr)
def pk(env, expr):
    return get_primary_key_type(env.context, env.backend)


@ufunc.resolver(Alembic, Expr)
def add_column(env, expr):
    args, kwargs = expr.resolve(env)
    return expr(*args, **kwargs)


@ufunc.executor(Alembic, Expr)
def add_column(env, expr):
    env.op.add_column(*expr.args, **expr.kwargs)


@ufunc.resolver(Alembic, Expr)
def drop_column(env, expr):
    args, kwargs = expr.resolve(env)
    return expr(*args, **kwargs)


@ufunc.executor(Alembic, Expr)
def drop_column(env, expr):
    env.op.drop_column(*expr.args, **expr.kwargs)


@ufunc.resolver(Alembic, Expr)
def alter_column(env, expr):
    args, kwargs = expr.resolve(env)
    table, name, *args = args
    if isinstance(table, Bind):
        table = table.name
    return Expr(expr.name, table, name, *args, **kwargs)


@ufunc.executor(Alembic, Expr)
def alter_column(env, expr):
    env.op.alter_column(*expr.args, **expr.kwargs)


@ufunc.resolver(Alembic, Expr)
def create_table(env, expr):
    args, kwargs = expr.resolve(env)
    table, *args = args
    if isinstance(table, Bind):
        table = table.name
    return Expr(expr.name, table, *args, **kwargs)


@ufunc.executor(Alembic, Expr)
def create_table(env, expr):
    env.op.create_table(*expr.args, **expr.kwargs)


@ufunc.resolver(Alembic, str)
def drop_table(env, table):
    return Expr('drop_table', table)


@ufunc.resolver(Alembic, Bind)
def drop_table(env, table):
    return Expr('drop_table', table.name)


@ufunc.executor(Alembic, Expr)
def drop_table(env, expr):
    env.op.drop_table(*expr.args, **expr.kwargs)


@ufunc.resolver(Alembic, Expr)
def column(env, expr):
    args, kwargs = expr.resolve(env)
    name, type_, *args = args
    if isinstance(name, Bind):
        name = name.name
    return sa.Column(name, type_, *args, **kwargs)


@ufunc.resolver(Alembic, Expr, names=list(TYPES))
def type_(env, name, expr):
    args, kwargs = expr.resolve(env)
    Type = TYPES[expr.name]
    return Type(*args, **kwargs)


@ufunc.resolver(Alembic, str)
def ref(env, reference, **kwargs):
    return sa.ForeignKey(reference, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spinta',
 'spinta.accesslog',
 'spinta.backends',
 'spinta.backends.fs',
 'spinta.backends.fs.commands',
 'spinta.backends.memory',
 'spinta.backends.memory.commands',
 'spinta.backends.mongo',
 'spinta.backends.mongo.commands',
 'spinta.backends.mongo.types',
 'spinta.backends.mongo.types.array',
 'spinta.backends.mongo.types.file',
 'spinta.backends.mongo.types.object',
 'spinta.backends.postgresql',
 'spinta.backends.postgresql.commands',
 'spinta.backends.postgresql.helpers',
 'spinta.backends.postgresql.types',
 'spinta.backends.postgresql.types.array',
 'spinta.backends.postgresql.types.file',
 'spinta.backends.postgresql.types.geometry',
 'spinta.backends.postgresql.types.object',
 'spinta.backends.postgresql.types.ref',
 'spinta.backends.types',
 'spinta.cli',
 'spinta.cli.helpers',
 'spinta.commands',
 'spinta.core',
 'spinta.datasets',
 'spinta.datasets.backends',
 'spinta.datasets.backends.csv',
 'spinta.datasets.backends.frictionless',
 'spinta.datasets.backends.frictionless.commands',
 'spinta.datasets.backends.notimpl',
 'spinta.datasets.backends.notimpl.commands',
 'spinta.datasets.backends.sql',
 'spinta.datasets.backends.sql.commands',
 'spinta.datasets.backends.sql.ufuncs',
 'spinta.datasets.backends.sqldump',
 'spinta.datasets.backends.sqldump.commands',
 'spinta.datasets.backends.sqldump.ufuncs',
 'spinta.datasets.commands',
 'spinta.datasets.keymaps',
 'spinta.dimensions',
 'spinta.dimensions.enum',
 'spinta.dimensions.lang',
 'spinta.dimensions.param',
 'spinta.dimensions.prefix',
 'spinta.dimensions.prefix.commands',
 'spinta.formats',
 'spinta.formats.ascii',
 'spinta.formats.csv',
 'spinta.formats.html',
 'spinta.formats.json',
 'spinta.formats.jsonlines',
 'spinta.hacks',
 'spinta.manifests',
 'spinta.manifests.backend',
 'spinta.manifests.backend.commands',
 'spinta.manifests.commands',
 'spinta.manifests.internal',
 'spinta.manifests.internal.commands',
 'spinta.manifests.tabular',
 'spinta.manifests.tabular.commands',
 'spinta.manifests.tabular.formats',
 'spinta.manifests.yaml',
 'spinta.manifests.yaml.commands',
 'spinta.migrations',
 'spinta.migrations.schema',
 'spinta.naming',
 'spinta.testing',
 'spinta.types',
 'spinta.types.array',
 'spinta.types.file',
 'spinta.types.geometry',
 'spinta.types.object',
 'spinta.types.ref',
 'spinta.ufuncs',
 'spinta.ufuncs.changebase',
 'spinta.units',
 'spinta.utils']

package_data = \
{'': ['*'], 'spinta': ['manifest/*', 'manifest/_schema/*', 'templates/*']}

install_requires = \
['GeoAlchemy2>=0.9.4,<0.10.0',
 'Shapely>=1.8.0,<2.0.0',
 'XlsxWriter>=1.3.7,<2.0.0',
 'aiofiles',
 'alembic',
 'asyncpg>=0.22.0,<0.23.0',
 'authlib>=0.11,<0.12',
 'click',
 'frictionless[sql]>=3.48.0,<4.0.0',
 'gunicorn',
 'jinja2',
 'jsonpatch',
 'lark-parser',
 'lxml',
 'msgpack',
 'multipledispatch',
 'openpyxl>=3.0.7,<4.0.0',
 'phonenumbers',
 'pprintpp',
 'psycopg2-binary',
 'pymongo',
 'python-multipart',
 'pytz',
 'requests',
 'ruamel.yaml',
 'setuptools-scm',
 'setuptools>=57,<58',
 'sqlalchemy>=1.4,<1.4.23',
 'sqlparse',
 'starlette',
 'texttable',
 'toposort',
 'tqdm',
 'typer[all]>=0.3.2,<0.4.0',
 'ujson',
 'unidecode',
 'uvicorn',
 'xlrd']

entry_points = \
{'console_scripts': ['spinta = spinta.cli.main:app']}

setup_kwargs = {
    'name': 'spinta',
    'version': '0.1.25',
    'description': 'A platform for describing, extracting, transforming, loading and serving open data.',
    'long_description': '.. default-role:: literal\n\nSpinta is a framework to describe, extract and publish data (DEP Framework). It\nsupports a great deal of data schemes and formats.\n\n.. image:: https://gitlab.com/atviriduomenys/spinta/badges/master/pipeline.svg\n   :target: https://gitlab.com/atviriduomenys/spinta/commits/master\n\n.. image:: https://gitlab.com/atviriduomenys/spinta/badges/master/coverage.svg\n   :target: https://gitlab.com/atviriduomenys/spinta/commits/master\n\n\nPurpose\n=======\n\n- **Describe your data**: You can automatically generate data structure\n  description table (*Manifest*) from many different data sources.\n\n- **Extract your data**: Once you have your data structure in *Manifest* tables,\n  you can extract data from multiple external data sources. Extracted data are\n  validated and transformed using rules defined in *Manifest* table. Finally,\n  data can be stored into internal database in order to provide fast and\n  flexible access to data.\n\n- **Publish your data**: Once you have your data loaded into internal\n  database, you can publish data using API. API is generated automatically using\n  *Manifest* tables and provides extracted data in many different formats. For\n  example if original data source was a CSV file, now you have a flexible API,\n  that can talk JSON, RDF, SQL, CSV and other formats.\n\n\nFeatures\n========\n\n- Simple 15 column table format for describing data structures (you can use\n  any spreadsheet software to manage metadata of your data)\n\n- Internal data storage with pluggable backends (PostgreSQL or Mongo)\n\n- Build-in async API server built on top of Starlette_ for data publishing\n\n- Simple web based data browser.\n\n- Convenient command-line interface\n\n- Public or restricted API access via OAuth protocol using build-in access\n  management.\n\n- Simple DSL_ for querying, transforming and validating data.\n\n- Low memory consumption for data of any size\n\n- Support for many different data sources\n\n- Advanced data extraction even from dynamic API.\n\n- Compatible with DCAT_ and `Frictionless Data Specifications`_.\n\n.. _Starlette: https://www.starlette.io/\n.. _DSL: https://en.wikipedia.org/wiki/Domain-specific_language\n.. _DCAT: https://www.w3.org/TR/vocab-dcat-2/\n.. _Frictionless Data Specifications: https://specs.frictionlessdata.io/\n\n\n.. note::\n\n    Currently this project is under heavy development and is in pre-alpha stage.\n    So many things might not work and many things can change. Use it at your own\n    risk.\n\n\nExample\n=======\n\nIf you have an SQLite database:\n\n.. code-block:: sh\n\n    $ sqlite3 sqlite.db <<EOF\n    CREATE TABLE COUNTRY (\n        NAME TEXT\n    );\n    EOF\n\nYou can get a limited API and simple web based data browser with a single\ncommand:\n\n.. code-block:: sh\n\n    $ spinta run -r sql sqlite:///sqlite.db\n\nThen you can generate metadata table (I call it *manifest*) like this:\n\n.. code-block:: sh\n\n    $ spinta inspect -r sql sqlite:///sqlite.db\n    d | r | b | m | property | type   | ref | source              | prepare | level | access | uri | title | description\n    dataset                  |        |     |                     |         |       |        |     |       |\n      | sql                  | sql    |     | sqlite:///sqlite.db |         |       |        |     |       |\n                             |        |     |                     |         |       |        |     |       |\n      |   |   | Country      |        |     | COUNTRY             |         |       |        |     |       |\n      |   |   |   | name     | string |     | NAME                |         | 3     | open   |     |       |\n\nGenerated data structure table can be saved into a CSV file:\n\n.. code-block:: sh\n\n    $ spinta inspect -r sql sqlite:///sqlite.db -o manifest.csv\n\nMissing peaces in metadata can be filled using any Spreadsheet software.\n\nOnce you done editing metadata, you can test it via web based data browser or\nAPI:\n\n.. code-block:: sh\n\n    $ spinta run --mode external manifest.csv\n\nOnce you are satisfied with metadata, you can generate a new metadata table for\npublishing, removing all traces of original data source:\n\n.. code-block:: sh\n\n    $ spinta copy --no-source --access open manifest.csv manifest-public.csv\n\nNow you have matadata for publishing, but all things about original data\nsource are gone. In order to publish data, you need to copy external data to\ninternal data store. To do that, first you need to initialize internal data\nstore:\n\n.. code-block:: sh\n\n    $ spinta config add backend my_backend postgresql postgresql://localhost/db\n    $ spinta config add manifest my_manifest tabular manifest-public.csv\n    $ spinta migrate\n\nOnce internal database is initialized, you can push external data into it:\n\n.. code-block:: sh\n\n    $ spinta push --access open manifest.csv\n\nAnd now you can publish data via full featured API with a web based data\nbrowser:\n\n.. code-block:: json\n\n    $ spinta run\n\nYou can access your data like this:\n\n.. code-block:: json\n\n    $ http :8000/dataset/sql/Country\n    HTTP/1.1 200 OK\n    content-type: application/json\n\n    {\n        "_data": [\n            {\n                "_type": "dataset/sql/Country",\n                "_id": "abdd1245-bbf9-4085-9366-f11c0f737c1d",\n                "_rev": "16dabe62-61e9-4549-a6bd-07cecfbc3508",\n                "_txn": "792a5029-63c9-4c07-995c-cbc063aaac2c",\n                "name": "Vilnius"\n            }\n        ]\n    }\n\n    $ http :8000/dataset/sql/Country/abdd1245-bbf9-4085-9366-f11c0f737c1d\n    HTTP/1.1 200 OK\n    content-type: application/json\n\n    {\n        "_type": "dataset/sql/Country",\n        "_id": "abdd1245-bbf9-4085-9366-f11c0f737c1d",\n        "_rev": "16dabe62-61e9-4549-a6bd-07cecfbc3508",\n        "_txn": "792a5029-63c9-4c07-995c-cbc063aaac2c",\n        "name": "Vilnius"\n    }\n\n    $ http :8000/dataset/sql/Country/abdd1245-bbf9-4085-9366-f11c0f737c1d?select(name)\n    HTTP/1.1 200 OK\n    content-type: application/json\n\n    {\n        "name": "Vilnius"\n    }\n',
    'author': 'Mantas Zimnickas',
    'author_email': 'sirexas@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/atviriduomenys/spinta',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import sqlalchemy as sa

from spinta.utils.schema import NA
from spinta.components import Context, Action, Property
from spinta.components import Context, Property, Action, DataItem
from spinta.backends.postgresql.components import PostgreSQL
from spinta.exceptions import UniqueConstraint


def pg_check_unique_constraint(
    context: Context,
    backend: PostgreSQL,
    prop: Property,
    name: str,
    data: DataItem,
    value: str,
):
    table = backend.get_table(prop)

    if (
        data.action in (Action.UPDATE, Action.PATCH) or
        data.action == Action.UPSERT and data.saved is not NA
    ):
        if prop.name == '_id' and value == data.saved['_id']:
            return
        condition = sa.and_(
            table.c[name] == value,
            table.c._id != data.saved['_id'],
        )
    else:
        condition = table.c[name] == value
    connection = context.get('transaction').connection
    default = object()
    result = backend.get(connection, table.c[name], condition, default=default)
    if result is not default:
        raise UniqueConstraint(prop)



from pkg_resources import get_distribution, DistributionNotFound

from setuptools_scm import get_version as get_scm_version

from spinta.commands import get_version
from spinta.components import Context


@get_version.register(Context)
def get_version(context: Context):
    try:
        version = get_distribution('spinta').version
    except DistributionNotFound:
        version = get_scm_version()

    return {
        'api': {
            'version': '0.0.1'
        },
        'implementation': {
            'name': 'Spinta',
            'version': version,
        },
        'build': None,
    }



from typing import List, Union

import sqlalchemy as sa

from sqlalchemy.sql.type_api import TypeEngine

from spinta import commands
from spinta.components import Context, Property
from spinta.types.datatype import Ref
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name, get_table_name
from spinta.backends.postgresql.helpers import get_column_name


@commands.prepare.register(Context, PostgreSQL, Ref)
def prepare(context: Context, backend: PostgreSQL, dtype: Ref):
    pkey_type = commands.get_primary_key_type(context, backend)
    return get_pg_foreign_key(
        dtype.prop,
        table_name=get_pg_name(get_table_name(dtype.model)),
        model_name=dtype.prop.model.name,
        column_type=pkey_type,
    )


def get_pg_foreign_key(
    prop: Property,
    *,
    table_name: str,
    model_name: str,
    column_type: TypeEngine,
) -> List[Union[sa.Column, sa.Constraint]]:
    column_name = get_column_name(prop) + '._id'
    return [
        sa.Column(column_name, column_type),
        sa.ForeignKeyConstraint(
            [column_name], [f'{table_name}._id'],
            name=get_pg_name(f'fk_{model_name}_{column_name}'),
        )
    ]



DATASET = [
    'id',
    'dataset',
    'resource',
    'base',
    'model',
    'property',
    'type',
    'ref',
    'source',
    'prepare',
    'level',
    'access',
    'uri',
    'title',
    'description',
]
